# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import Callable, Optional
import traceback

# Third Party
from grpc._utilities import RpcMethodHandler
from prometheus_client import Gauge
import grpc

# First Party
import alog

# Local
from caikit.runtime.service_factory import ServicePackage
from caikit.runtime.service_generation.rpcs import CaikitRPCBase
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException

log = alog.use_channel("SERVER-WRAPR")

IN_PROGRESS_GAUGE = Gauge(
    "rpc_in_progress_gauge",
    "Total number of in-flight requests to caikit-runtime",
    ["rpc_name"],
)


class CaikitRuntimeServerWrapper(grpc.Server):
    """This class wraps an underlying gRPC server for the purpose of
    intercepting the binding of servicers (e.g., the CaikitRuntimeServicer) to
    the server so that the RPC handlers that are registered to the server
    can optionally be replaced with a generic global predict RPC handler
    instead.
    """

    def __init__(self, server, global_predict, intercepted_svc_package: ServicePackage):
        """Initialize a new CaikitRuntimeServerWrapper

        Args:
            server(grpc.Server): The server that is being wrapped
            global_predict(function): A function that will accept an arbitrary
                gRPC request message and a grpc.ServicerContext, and return
                a suitable gRPC response message
        """

        self._server = server
        self._global_predict = global_predict
        self._intercepted_svc_package = intercepted_svc_package
        self._intercepted_methods = []

        for method in self._intercepted_svc_package.descriptor.methods:
            # Take the method short name (e.g., 'SyntaxIzumoPredict') and
            # concatenate it with the intercepted service name to produce
            # a fully qualified RPC method name that we wish to intercept
            # (e.g., '/natural_language_understanding.CaikitRuntime/SyntaxIzumoPredict')
            fqm = "/%s/%s" % (
                self._intercepted_svc_package.descriptor.full_name,
                method.name,
            )

            log.info("<RUN81194024I>", "Intercepting RPC method %s", fqm)
            self._intercepted_methods.append((method.name, fqm))

    # **************************************************************************
    # Custom methods
    # **************************************************************************

    def intercepted_service(self):
        """Get the fully-qualified name of the intercepted service

        Returns:
            string:
                The fully-qualified name of the service whose RPC handlers are
                intercepted by this server wrapper
        """
        return self._intercepted_svc_package.descriptor.full_name

    def intercepted_methods(self):
        """Get the list of intercepted predict RPC methods

        Returns:
            list((string, string)):
                A list of two-element tuples containing the short name (e.g.,
                'SyntaxIzumoPredict') and fully-qualified name (e.g.,
                '/natural_language_understanding.CaikitRuntime/SyntaxIzumoPredict')
                of every RPC method intercepted by this server wrapper
        """
        return self._intercepted_methods

    @staticmethod
    def safe_rpc_wrapper(rpc: Callable, caikit_rpc: Optional[CaikitRPCBase] = None):
        """This wrapper should be used to safely invoke an RPC. If used, it adds automatic error
        handling and conversion to the appropriate response for gRPC, as well as logging indicating
        if the the error was intentional (i.e., thrown as CaikitRuntimeException directly) or
        unexpected (i.e., thrown as a non GRPC error).

        Args:
            rpc(Function): Method attached to a servicer instance to be invoked in a safe manner.

        Returns:
            A function that takes a gRPC request message and ServicerContext, and safely invokes
            the provided RPC.
        """
        if rpc is None:
            message = "Programming error, RPC is None!"
            log.error("<RUN33322123E>", message)
            raise CaikitRuntimeException(grpc.StatusCode.INTERNAL, message)

        if rpc.__name__ == "safe_rpc_call":
            return rpc
        log.info(
            "<RUN33333123I>",
            "Wrapping safe rpc for %s",
            rpc.__name__,
        )

        def safe_rpc_call(request, context):
            """This function should be used to safely invoke an RPC. If used, it adds automatic
            error handling and conversion to the appropriate response for gRPC, as well as logging
            indicating if the the error was intentional (i.e., thrown as CaikitRuntimeException
            directly) or unexpected (i.e., thrown as a non GRPC error).

            Args:
                request(message.Message): gRPC request object normally received by the rpc.
                context(grpc.ServicerContext): gRPC context object normally received by the rpc.

            Returns:
                gRPC response object return by the invoked RPC or None (aborted context).
            """
            with alog.ContextLog(log.debug, "[Safe RPC]: %s", rpc.__name__):
                try:
                    IN_PROGRESS_GAUGE.labels(rpc_name=rpc.__name__).inc()
                    if caikit_rpc:
                        # Pass through the CaikitRPCBase rpc description to the global handlers
                        return rpc(request, context, caikit_rpc=caikit_rpc)
                    return rpc(request, context)

                except CaikitRuntimeException as e:
                    log_dict = {"log_code": "<RUN89011375W>", "message": e.message}
                    log.warning({**log_dict, **e.metadata})
                    context.abort(e.status_code, e.message)

                except ValueError as e:
                    message = repr(e)
                    log.error("<RUN33333333E>", message)
                    log.error("<RUN33333334E>", str(traceback.format_exc()))
                    context.abort(grpc.StatusCode.UNKNOWN, message)
                finally:
                    IN_PROGRESS_GAUGE.labels(rpc_name=rpc.__name__).dec()

        return safe_rpc_call

    # **************************************************************************
    # Overridden grpc.Server methods
    # **************************************************************************

    def add_generic_rpc_handlers(self, generic_rpc_handlers):
        """Registers GenericRpcHandlers with this Server.

        This method will intercept the generic_rpc_handlers

        Args:
          generic_rpc_handlers: An iterable of GenericRpcHandlers that will be
          used to service RPCs.
        """

        class DummyHandlerCallDetails(grpc.HandlerCallDetails):
            """Dummy class for constructing a grpc.HandlerCallDetails object"""

            def __init__(self, method):
                super().__init__()
                self.method = method

        # Iterate over each grpc.ServiceRpcHandler...
        for handler in generic_rpc_handlers:
            # ...and check if this is the service we wish to intercept
            if handler.service_name() == self.intercepted_service():
                # This is the service whose RPC handlers we wish to intercept
                # and re-route.  We now need to iterate over each method that
                # we wish to re-route, get the original RPC handler for that
                # method (see
                # caikit_runtime_pb2_grpc.add_CaikitRuntimeServicer_to_server
                # for a dict of the rpc_method_handlers we wish to re-route)
                rerouted_rpc_method_handlers = {}
                for method, fqm in self.intercepted_methods():
                    # Get the original grpc.RpcMethodHandler for this RPC method
                    original_rpc_handler = handler.service(DummyHandlerCallDetails(fqm))

                    # Find the Caikit RPC that maps to this rpc
                    caikit_rpc = self._intercepted_svc_package.caikit_rpcs.get(
                        method, None
                    )
                    if not caikit_rpc:
                        raise ValueError(f"No Caikit RPC Found for method: {method}")

                    # Now, swap out the original unary-unary callable with our
                    # generic predict method, and add this newly re-routed RPC
                    # method handler to the dict of (method, handler) pairs
                    safe_rpc_handler = self._make_new_handler(
                        original_rpc_handler, caikit_rpc
                    )
                    rerouted_rpc_method_handlers[method] = safe_rpc_handler
                    log.info(
                        "<RUN30032825I>",
                        "Re-routing RPC %s from %s to %s",
                        fqm,
                        self._get_handler_fn(original_rpc_handler),
                        self._get_handler_fn(rerouted_rpc_method_handlers[method]),
                    )

                # Now that we have re-rerouted all the original RPC method
                # handlers to the global predict RPC method handler, it is time
                # to bind them to the underlying server that we are wrapping
                generic_handler = grpc.method_handlers_generic_handler(
                    self.intercepted_service(), rerouted_rpc_method_handlers
                )
                self._server.add_generic_rpc_handlers((generic_handler,))
                log.info(
                    "<RUN24924908I>",
                    "Interception of service %s complete",
                    self.intercepted_service(),
                )

            else:
                # This is not the service whose RPC handlers we wish to
                # intercept, so just pass the (unmodified) RPC handlers
                # along to the underlying gRPC server we are wrapping
                assert isinstance(handler, grpc._utilities.DictionaryGenericHandler)
                for method in handler._method_handlers:
                    # Wrap the RPC handler for this method in a safe RPC call,
                    # but do not replace the handler with a global handler
                    original_rpc_handler = handler._method_handlers[method]
                    safe_rpc_handler = self._make_new_handler(original_rpc_handler)
                    handler._method_handlers[method] = safe_rpc_handler

                self._server.add_generic_rpc_handlers(generic_rpc_handlers)

    def _make_new_handler(
        self,
        original_rpc_handler: RpcMethodHandler,
        caikit_rpc: Optional[CaikitRPCBase] = None,
    ):
        if caikit_rpc:
            behavior = self.safe_rpc_wrapper(self._global_predict, caikit_rpc)
        else:
            behavior = self.safe_rpc_wrapper(self._get_handler_fn(original_rpc_handler))

        if original_rpc_handler.unary_unary:
            handler_constructor = grpc.unary_unary_rpc_method_handler
        elif original_rpc_handler.unary_stream:
            handler_constructor = grpc.unary_stream_rpc_method_handler
        elif original_rpc_handler.stream_unary:
            handler_constructor = grpc.stream_unary_rpc_method_handler
        else:
            handler_constructor = grpc.stream_stream_rpc_method_handler

        return handler_constructor(
            behavior=behavior,
            request_deserializer=original_rpc_handler.request_deserializer,
            response_serializer=original_rpc_handler.response_serializer,
        )

    # **************************************************************************
    # Pass-through (i.e., unchanged) grpc.Server methods
    # **************************************************************************

    def add_insecure_port(self, address):
        """Opens an insecure port for accepting RPCs.

        This method may only be called before starting the server.

        Args:
          address: The address for which to open a port.
          if the port is 0, or not specified in the address, then gRPC runtime
          will choose a port.

        Returns:
          integer:
          An integer port on which server will accept RPC requests.
        """
        return self._server.add_insecure_port(address)

    def add_secure_port(self, address, server_credentials):
        """Opens a secure port for accepting RPCs.

        This method may only be called before starting the server.

        Args:
          address: The address for which to open a port.
            if the port is 0, or not specified in the address, then gRPC
            runtime will choose a port.
          server_credentials: A ServerCredentials object.
        Returns:
          integer:
          An integer port on which server will accept RPC requests.
        """
        return self._server.add_secure_port(address, server_credentials)

    def start(self):
        """Starts this Server.

        This method may only be called once. (i.e. it is not idempotent).
        """
        self._server.start()

    def stop(self, grace):
        """Stops this Server.

        This method immediately stop service of new RPCs in all cases.

        If a grace period is specified, this method returns immediately
        and all RPCs active at the end of the grace period are aborted.
        If a grace period is not specified (by passing None for `grace`),
        all existing RPCs are aborted immediately and this method
        blocks until the last RPC handler terminates.

        This method is idempotent and may be called at any time.
        Passing a smaller grace value in a subsequent call will have
        the effect of stopping the Server sooner (passing None will
        have the effect of stopping the server immediately). Passing
        a larger grace value in a subsequent call *will not* have the
        effect of stopping the server later (i.e. the most restrictive
        grace value is used).

        Args:
          grace: A duration of time in seconds or None.
        Returns:
          A threading.Event that will be set when this Server has completely
          stopped, i.e. when running RPCs either complete or are aborted and
          all handlers have terminated.
        """
        return self._server.stop(grace)

    def wait_for_termination(self, timeout=None):
        """Block current thread until the server stops.
        The wait will not consume computational resources during blocking,
        and it will block until one of the two following conditions are met:
        1. The server is stopped or terminated;
        2. A timeout occurs if timeout is not None.

        The timeout argument works in the same way as threading.Event.wait().
        Args:
            timeout: A floating point number specifying a timeout for the operation in seconds.
        Returns:
            A bool indicates if the operation times out.
        """
        return self._server.wait_for_termination(timeout)

    @staticmethod
    def _get_handler_fn(handler: RpcMethodHandler) -> Callable:
        if handler.unary_unary:
            return handler.unary_unary
        if handler.unary_stream:
            return handler.unary_stream
        if handler.stream_unary:
            return handler.stream_unary
        return handler.stream_stream



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Standard
import importlib
import sys

# First Party
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("COMPATIBILITY")
error = error_handler.get(log)


def unsupported_platforms(platforms):
    """This decorator can be used to raise NotImplementedError if it's run on a given platform.
    Otherwise it runs as normal. For guidance on different permissible values of platforms,
    see: https://docs.python.org/3/library/sys.html#sys.platform

    Args:
        platforms (List | Tuple | str): Platforms on which this capability is
            not implemented.
    """
    error.type_check("<COR48137162E>", list, tuple, str, platforms=platforms)
    if not isinstance(platforms, str):
        error.type_check_all("<COR92928316E>", str, platforms=platforms)

    def decorator(func):
        def wrapper(*args, **kwargs):
            if (
                isinstance(platforms, str) and sys.platform == platforms
            ) or sys.platform in platforms:
                error(
                    "<COR12233351E>",
                    NotImplementedError(
                        "Operation [{}] has not been implemented for platform [{}]".format(
                            func.__name__, sys.platform
                        )
                    ),
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def is_importable_on_platform(module, package, exc_type, platforms, platform_hint):
    """Some modules have trouble on specific operating systems. This function takes
    a module and package from which we want to import, and an exception type to look
    out for on a given problem. If the exception is raised, we warn the platform hint.

    NOTE: This function does set any properties on your package. You should use this
    function to check if something is importable, then you should import it. This will
    only run the package initialization one time, because this function will add your
    module into the loaded packages if it's successful, but doing things this way prevents
    your package from having a bunch of undesirable attributes and keeps it cleaner than other
    approaches. It also ensures that the import stack trace doesn't look abnormally handled if we
    raise on other platforms.

    Args:
        module (str): Module that we want to import.
        package (str): Package from which we want to import the module.
        exc_type (Exception): Exception that we want to warn a hint for if we
            hit it on import for a given platform.
        platforms (str): Platform on which we expect potential bad import from
            for this module.
        platform_hint (str): Hint to warn if we hit the exception type on the
            provided platform.
    """
    error.type_check(
        "<COR48442162E>",
        str,
        module=module,
        package=package,
        platform_hint=platform_hint,
    )
    error.type_check("<COR62245359E>", list, tuple, str, platforms=platforms)
    if not isinstance(platforms, str):
        error.type_check_all("<COR86935489E>", str, platforms=platforms)
    # If this is not a platform that needs special handling, just return True
    if (
        isinstance(platforms, str) and sys.platform != platforms
    ) or sys.platform not in platforms:
        return True
    # We are running on a platform that has potential issues for this module.
    # If we succeed on import, return True. If we throw the known exception type,
    # catch it and log the hint, then return False. This is non-fatal.
    try:
        log.debug(
            "Determining if module {} is importable on platform {}".format(
                module, platforms
            )
        )
        importlib.import_module(module, package)
        return True
    except exc_type as e:
        log.warning(
            "Module [{}] from package [{}] failed to import on platform [{}]"
            " with error [{}] and will not be available.\nHINT: {}".format(
                module, package, platforms, str(e), platform_hint
            )
        )
        return False



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for DictBackend"""

# Standard
from typing import Iterable

# Third Party
import pytest

# Local
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.data_backends.dict_backend import DictBackend
from tests.data_model_helpers import make_proto_def, temp_data_model


def test_dict_backend_basic_message():
    """Test that the dict backend can be used for messages containing primitives"""
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": str,
                    "bar": int,
                }
            }
        )
    ) as dm:
        assert hasattr(dm, "ThingOne")
        assert issubclass(dm.ThingOne, DataBase)

        data_dict = {
            "foo": "foo-val",
            "bar": 1234,
        }

        backend = DictBackend(data_dict)
        msg = dm.ThingOne.from_backend(backend)

        assert msg.foo == data_dict["foo"]
        assert msg.bar == data_dict["bar"]


def test_dict_backend_sub_message():
    """Make sure that a nested message is handled correctly"""
    with temp_data_model(
        make_proto_def(
            {
                "SomeThing": {
                    "foo": str,
                    "bar": int,
                },
                "ThingWrapper": {
                    "the_thing": "SomeThing",
                },
            }
        )
    ) as dm:
        assert hasattr(dm, "SomeThing")
        assert hasattr(dm, "ThingWrapper")
        assert issubclass(dm.SomeThing, DataBase)
        assert issubclass(dm.ThingWrapper, DataBase)

        data_dict = {
            "the_thing": {
                "foo": "foo-val",
                "bar": 1234,
            },
        }

        backend = DictBackend(data_dict)
        msg = dm.ThingWrapper.from_backend(backend)

        sub_msg = msg.the_thing
        assert isinstance(sub_msg, DataBase)
        assert sub_msg.foo == data_dict["the_thing"]["foo"]
        assert sub_msg.bar == data_dict["the_thing"]["bar"]


def test_dict_backend_repeated_sub_message():
    """Make sure that a nested repeated message is handled correctly"""
    with temp_data_model(
        make_proto_def(
            {
                "Foo": {
                    "foo": str,
                    "bar": int,
                },
                "FooWrapper": {
                    "some_foos": ["Foo"],
                },
            }
        )
    ) as dm:
        assert hasattr(dm, "Foo")
        assert hasattr(dm, "FooWrapper")
        assert issubclass(dm.Foo, DataBase)
        assert issubclass(dm.FooWrapper, DataBase)

        data_dict = {
            "some_foos": [
                {"foo": "foo-val", "bar": 1234},
                {"foo": "baz-val", "bar": 4321},
            ],
        }

        backend = DictBackend(data_dict)
        msg = dm.FooWrapper.from_backend(backend)
        sub_msgs = msg.some_foos
        assert isinstance(sub_msgs, Iterable)
        assert len(sub_msgs) == 2
        sub_msg1, sub_msg2 = sub_msgs
        assert sub_msg1.foo == data_dict["some_foos"][0]["foo"]
        assert sub_msg1.bar == data_dict["some_foos"][0]["bar"]
        assert sub_msg2.foo == data_dict["some_foos"][1]["foo"]
        assert sub_msg2.bar == data_dict["some_foos"][1]["bar"]


def test_dict_backend_oneof():
    """Make sure that a oneof can be correctly accessed from a backend"""
    with temp_data_model(make_proto_def({"Foo": {"foo": "Union[str, int]"}})) as dm:
        assert hasattr(dm, "Foo")
        assert issubclass(dm.Foo, DataBase)

        data_dict = {"foo": "asdf"}
        backend = DictBackend(data_dict)
        msg = dm.Foo.from_backend(backend)
        assert msg.foo == "asdf"
        assert msg.which_oneof("foo") == "foo_str"


def test_dict_backend_invalid_field_error():
    """Make sure that an AttributeError is raised if an invalid field is
    requested
    """
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": str,
                    "bar": int,
                }
            }
        )
    ) as dm:
        data_dict = {
            "foo": "foo-val",
            "bar": 1234,
        }

        backend = DictBackend(data_dict)
        msg = dm.ThingOne.from_backend(backend)

        with pytest.raises(AttributeError):
            backend.get_attribute(dm.ThingOne, "baz")

        with pytest.raises(AttributeError):
            msg.baz



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from concurrent import futures

# Third Party
import grpc

# Local
from caikit.core.data_model.base import DataBase
from caikit.runtime.interceptors.caikit_runtime_server_wrapper import (
    CaikitRuntimeServerWrapper,
)
from caikit.runtime.service_generation.rpcs import TaskPredictRPC
from sample_lib.data_model import SampleOutputType


def test_rpc_is_passed_to_predict_handlers(sample_inference_service, open_port):
    calls = []

    def predict(request, context, caikit_rpc):
        calls.append(caikit_rpc)
        return SampleOutputType().to_proto()

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10),
    )
    wrapper = CaikitRuntimeServerWrapper(server, predict, sample_inference_service)
    sample_inference_service.registration_function(
        sample_inference_service.service, wrapper
    )
    wrapper.add_insecure_port(f"[::]:{open_port}")

    try:
        wrapper.start()

        client = sample_inference_service.stub_class(
            grpc.insecure_channel(f"localhost:{open_port}")
        )
        predict_class = DataBase.get_class_for_name("SampleTaskRequest")
        _ = client.SampleTaskPredict(predict_class().to_proto(), timeout=3)
        assert len(calls) == 1
        assert isinstance(calls[0], TaskPredictRPC)
        assert calls[0].name == "SampleTaskPredict"
    finally:
        wrapper.stop(0)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Shared registry of known backend types
"""

# Standard
from typing import Optional, Type

# First Party
import alog

# Local
from ..exceptions import error_handler
from ..registries import module_backend_classes, module_backend_types
from ..toolkit.wip_decorator import Action, WipCategory, work_in_progress
from .base import BackendBase

log = alog.use_channel("BCKENDTYP")
error = error_handler.get(log)


## Public ######################################################################


@work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
def register_backend_type(config_class: Optional[Type[BackendBase]] = None):
    """Register a new module backend type by name. This will perform the
    registration if the value does not already exist in the shared enum and will
    be a no-op if it is already present. The backend name should be present as
    a property of the config_class

    Args:
        config_class (BackendBase): Module to configure particular backend
    """
    # Type validation
    error.type_check(
        "<COR72253432E>",
        type(BackendBase),
        allow_none=True,
        config_class=config_class,
    )

    # NOTE: if a config_class is of `BackendBase` then it will always have
    # backend_type as property which contains the name
    type_name = config_class.backend_type

    error.type_check("<COR92253963E>", str, type_name=type_name)

    # Case check
    error.value_check(
        "<COR07825070E>",
        type_name.upper() == type_name,
        "type_name must be fully UPPERCASE: {}",
        type_name,
    )

    # Add to the global registries
    # NOTE: This only contains a module name and does not contain an object of the backend module
    # The object of the "configured" backend module is generated using caikit.config.configure
    backend_types = module_backend_types()
    backend_classes = module_backend_classes()
    if type_name not in backend_types:
        backend_types[type_name] = type_name
        backend_classes[type_name] = config_class


def __getattr__(name):
    """This module forwards attribute access to the MODULE_BACKEND_TYPES
    mapping
    """
    if not name.startswith("_"):
        return getattr(module_backend_types(), name)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from typing import Iterator
from unittest.mock import MagicMock, patch

# Local
from caikit.runtime.service_factory import get_inference_request
from sample_lib.modules.geospatial import GeoStreamingModule

try:
    # Standard
    from test.support.threading_helper import catch_threading_exception
except (NameError, ModuleNotFoundError):
    from tests.base import catch_threading_exception

# Standard
import json
import threading
import time

# Third Party
import grpc
import pytest

# Local
from caikit.runtime.servicers.global_predict_servicer import GlobalPredictServicer
from caikit.runtime.types.aborted_exception import AbortedException
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from sample_lib.data_model import SampleInputType, SampleOutputType
from sample_lib.modules.sample_task import SampleModule
from tests.conftest import temp_config
from tests.fixtures import Fixtures

HAPPY_PATH_INPUT_DM = SampleInputType(name="Gabe")
HAPPY_PATH_RESPONSE_DM = SampleOutputType(greeting="Hello Gabe")
HAPPY_PATH_INPUT = HAPPY_PATH_INPUT_DM.to_proto()
HAPPY_PATH_RESPONSE = HAPPY_PATH_RESPONSE_DM.to_proto()


def test_calling_predict_should_raise_if_module_raises(
    sample_inference_service,
    sample_predict_servicer,
    sample_task_model_id,
    sample_task_unary_rpc,
):
    with pytest.raises(CaikitRuntimeException) as context:
        # SampleModules will raise a RuntimeError if the throw flag is set
        predict_class = get_inference_request(SampleModule.TASK_CLASS)
        request = predict_class(sample_input=HAPPY_PATH_INPUT_DM, throw=True).to_proto()
        sample_predict_servicer.Predict(
            request,
            Fixtures.build_context(sample_task_model_id),
            caikit_rpc=sample_task_unary_rpc,
        )
    assert context.value.status_code == grpc.StatusCode.INTERNAL
    assert "Unhandled exception during prediction" in context.value.message


def test_invalid_input_to_a_valid_caikit_core_class_method_raises(
    sample_task_model_id,
    sample_inference_service,
    sample_predict_servicer,
    sample_task_unary_rpc,
):
    """Test that a caikit.core module that gets an unexpected input value errors in an expected way"""
    with pytest.raises(CaikitRuntimeException) as context:
        # SampleModules will raise a ValueError if the poison pill name is given
        predict_class = get_inference_request(SampleModule.TASK_CLASS)
        request = predict_class(
            sample_input=SampleInputType(name=SampleModule.POISON_PILL_NAME)
        ).to_proto()
        sample_predict_servicer.Predict(
            request,
            Fixtures.build_context(sample_task_model_id),
            caikit_rpc=sample_task_unary_rpc,
        )
    assert context.value.status_code == grpc.StatusCode.INVALID_ARGUMENT
    assert "problem with your input" in context.value.message


def test_global_predict_works_for_unary_rpcs(
    sample_inference_service,
    sample_predict_servicer,
    sample_task_model_id,
    sample_task_unary_rpc,
):
    """Global predict of SampleTaskRequest returns a prediction"""
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    response = sample_predict_servicer.Predict(
        predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
        Fixtures.build_context(sample_task_model_id),
        caikit_rpc=sample_task_unary_rpc,
    )
    assert response == HAPPY_PATH_RESPONSE


def test_global_predict_works_on_bidirectional_streaming_rpcs(
    sample_inference_service, sample_predict_servicer, sample_task_model_id
):
    """Simple test that our SampleModule's bidirectional stream inference fn is supported"""

    predict_class = get_inference_request(
        SampleModule.TASK_CLASS, input_streaming=True, output_streaming=True
    )

    def req_iterator() -> Iterator[predict_class]:
        for i in range(100):
            yield predict_class(sample_inputs=HAPPY_PATH_INPUT_DM).to_proto()

    response_stream = sample_predict_servicer.Predict(
        req_iterator(),
        Fixtures.build_context(sample_task_model_id),
        caikit_rpc=sample_inference_service.caikit_rpcs[
            "BidiStreamingSampleTaskPredict"
        ],
    )
    count = 0
    for response in response_stream:
        assert response == HAPPY_PATH_RESPONSE
        count += 1
    assert count == 100


def test_global_predict_works_on_bidirectional_streaming_rpcs_with_multiple_streaming_parameters(
    sample_inference_service, sample_predict_servicer, sample_task_model_id
):
    """Test that our little geospatial model that takes multiple streams is supported"""

    mock_manager = MagicMock()
    mock_manager.retrieve_model.return_value = GeoStreamingModule()

    predict_class = get_inference_request(
        GeoStreamingModule.TASK_CLASS, input_streaming=True, output_streaming=True
    )

    def req_iterator() -> Iterator[predict_class]:
        for i in range(100):
            yield predict_class(lats=i, lons=100 - i, name="Gabe").to_proto()

    with patch.object(sample_predict_servicer, "_model_manager", mock_manager):
        response_stream = sample_predict_servicer.Predict(
            req_iterator(),
            Fixtures.build_context(sample_task_model_id),
            caikit_rpc=sample_inference_service.caikit_rpcs[
                "BidiStreamingGeoSpatialTaskPredict"
            ],
        )
        count = 0
        for i, response in enumerate(response_stream):
            assert response.greeting == f"Hello from Gabe at {i}.0°, {100-i}.0°"
            count += 1
        assert count == 100


def test_global_predict_predict_model_direct(
    sample_inference_service, sample_predict_servicer, sample_task_model_id
):
    """Test that the direct predict_model function can be called to bypass the
    gRPC handler
    """
    response = sample_predict_servicer.predict_model(
        request_name="SampleTaskRequest",
        model_id=sample_task_model_id,
        sample_input=HAPPY_PATH_INPUT_DM,
    )
    assert response == HAPPY_PATH_RESPONSE_DM


def test_global_predict_aborts_long_running_predicts(
    sample_inference_service,
    sample_predict_servicer,
    sample_task_unary_rpc,
):
    mock_manager = MagicMock()

    # return a dummy model from the mock model manager
    class UnresponsiveModel(SampleModule):
        started = threading.Event()

        def run(self, *args, **kwargs):
            self.started.set()
            while True:
                time.sleep(0.001)

    dummy_model = UnresponsiveModel()
    mock_manager.retrieve_model.return_value = dummy_model

    context = Fixtures.build_context("test-any-unresponsive-model")
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    predict_thread = threading.Thread(
        target=sample_predict_servicer.Predict,
        args=(
            predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
            context,
        ),
        kwargs={"caikit_rpc": sample_task_unary_rpc},
    )

    with catch_threading_exception() as cm:
        # Patch in the mock manager and start the prediction
        with patch.object(sample_predict_servicer, "_model_manager", mock_manager):
            predict_thread.start()
            assert dummy_model.started.wait(2)
            # Simulate a timeout or client abort
            context.cancel()
            predict_thread.join(10)

        # Make sure the prediction actually stopped
        assert not predict_thread.is_alive()

        # Make sure the correct exception was raised
        assert cm.exc_type == AbortedException


def test_metering_ignore_unsuccessful_calls(
    sample_inference_service,
    sample_task_model_id,
    sample_task_unary_rpc,
):
    with temp_config({"runtime": {"metering": {"enabled": True}}}, "merge"):
        gps = GlobalPredictServicer(sample_inference_service)
        try:
            with patch.object(gps.rpc_meter, "update_metrics") as mock_update_func:
                predict_class = get_inference_request(SampleModule.TASK_CLASS)
                request = predict_class(
                    sample_input=HAPPY_PATH_INPUT_DM, throw=True
                ).to_proto()
                with pytest.raises(CaikitRuntimeException):
                    gps.Predict(
                        request,
                        Fixtures.build_context(sample_task_model_id),
                        caikit_rpc=sample_task_unary_rpc,
                    )

                mock_update_func.assert_not_called()
        finally:
            gps.stop_metering()


def test_metering_predict_rpc_counter(
    sample_inference_service, sample_task_model_id, sample_task_unary_rpc
):
    # need a new servicer to get a fresh new RPC meter
    with temp_config({"runtime": {"metering": {"enabled": True}}}, "merge"):
        sample_predict_servicer = GlobalPredictServicer(sample_inference_service)
        try:
            # Making 20 requests
            predict_class = get_inference_request(SampleModule.TASK_CLASS)
            for i in range(20):
                sample_predict_servicer.Predict(
                    predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
                    Fixtures.build_context(sample_task_model_id),
                    caikit_rpc=sample_task_unary_rpc,
                )

            # Force meter to write
            sample_predict_servicer.rpc_meter.flush_metrics()

            # Assertions on the created metrics file
            with open(sample_predict_servicer.rpc_meter.file_path) as f:
                data = [json.loads(line) for line in f]

            assert len(data) == 1
            assert list(data[0].keys()) == [
                "timestamp",
                "batch_size",
                "model_type_counters",
                "container_id",
            ]
            assert data[0]["batch_size"] == 20
            assert len(data[0]["model_type_counters"]) == 1
            assert data[0]["model_type_counters"] == {
                "<class 'sample_lib.modules.sample_task.sample_implementation.SampleModule'>": 20
            }
        finally:
            sample_predict_servicer.stop_metering()


def test_metering_write_to_metrics_file_twice(
    sample_inference_service,
    sample_task_model_id,
    sample_task_unary_rpc,
):
    """Make sure subsequent metering events append to file"""
    with temp_config({"runtime": {"metering": {"enabled": True}}}, "merge"):
        # need a new servicer to get a fresh new RPC meter
        sample_predict_servicer = GlobalPredictServicer(sample_inference_service)
        try:
            predict_class = get_inference_request(SampleModule.TASK_CLASS)
            sample_predict_servicer.Predict(
                predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
                Fixtures.build_context(sample_task_model_id),
                caikit_rpc=sample_task_unary_rpc,
            )

            # Force write
            sample_predict_servicer.rpc_meter.flush_metrics()

            sample_predict_servicer.Predict(
                predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
                Fixtures.build_context(sample_task_model_id),
                caikit_rpc=sample_task_unary_rpc,
            )

            # Force write
            sample_predict_servicer.rpc_meter.flush_metrics()

            with open(sample_predict_servicer.rpc_meter.file_path) as f:
                data = [json.loads(line) for line in f]

            assert len(data) == 2
            assert list(data[0].keys()) == [
                "timestamp",
                "batch_size",
                "model_type_counters",
                "container_id",
            ]
            assert data[0]["batch_size"] == 1
            assert len(data[0]["model_type_counters"]) == 1
            assert data[0]["model_type_counters"] == {
                "<class 'sample_lib.modules.sample_task.sample_implementation.SampleModule'>": 1
            }
        finally:
            sample_predict_servicer.stop_metering()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package handles all the gorp of finding types from docstrings given our custom
conventions"""

# Standard
from typing import Callable, Dict, List, Optional, Tuple, Type, Union
import builtins
import sys

# Third Party
from docstring_parser import ParseError
import docstring_parser

# First Party
import alog

# Local
from caikit.core.data_model.base import DataBase
import caikit.core

log = alog.use_channel("DOCSTRINGS")


def get_return_type(fn: Callable) -> Optional[Type]:
    """
    Grabs the return type off the docstring, if possible
    Args:
        fn: The function to get the return value of
            e.g. my_caikit_library.modules.classification.Transformer.run

    Returns:
        The return type of `fn`, if it can be parsed from the docstring. Otherwise, None
    """
    try:
        docstring = docstring_parser.parse(fn.__doc__)
    except docstring_parser.ParseError as e:
        log.warning("Could not parse docstring: %s fn.__doc__ ", fn.__doc__, exc_info=e)
        return None

    type_names, desc_names = _get_candidate_type_names_from_docstring(docstring.returns)

    return_type = _get_docstring_type(type_names)
    if return_type:
        return return_type

    return _get_docstring_type(desc_names)


def is_optional(fn: Callable, arg_name: str) -> bool:
    """
    Checks if the `arg_name` param from `fn`s docstring is optional
    by checking if param description starts with "an optional"
    or "optional".

    Args:
        fn: The function to get the type of a parameter from
            e.g.  my_caikit_library.modules.classification.Transformer.run
        arg_name: The name of the parameter that we should try to get the type of
            e.g. "raw_document"
    """
    try:
        docstring = docstring_parser.parse(fn.__doc__)
    except ParseError as parse_error:
        log.warning(
            "Failed to parse docstring for %s when looking for optional flag on parameter %s",
            fn,
            arg_name,
            exc_info=parse_error,
        )
        return False

    ds_param = [param for param in docstring.params if param.arg_name == arg_name]
    if ds_param:
        if len(ds_param) > 1:
            log.warning("Docstring has multiple args with the same name! %s", arg_name)
        ds_param = ds_param[0]

        if ds_param.description is not None:
            for description_line in ds_param.description.split("\n"):
                if description_line.lower().startswith(
                    "optional"
                ) or description_line.lower().startswith("an optional"):
                    log.debug2("Optional parameter found: %s", ds_param)
                    return True

        return False
    return False


def get_arg_type(fn: Callable, arg_name: str) -> Optional[Type]:
    """
    Grabs the type of the `arg_name` param from `fn`s docstring, if possible
    Args:
        fn: The function to get the type of a parameter from
            e.g. my_caikit_library.modules.classification.Transformer.run
        arg_name: The name of the parameter that we should try to get the type of
            e.g. "raw_document"
    Returns:
        The return type of `fn`, if it can be parsed from the docstring. Otherwise, None
    """
    try:
        docstring = docstring_parser.parse(fn.__doc__)
    except ParseError as parse_error:
        log.warning(
            "Failed to parse docstring for %s when looking for type on parameter %s",
            fn,
            arg_name,
            exc_info=parse_error,
        )
        return None

    ds_param = [param for param in docstring.params if param.arg_name == arg_name]
    if ds_param:
        if len(ds_param) > 1:
            log.warning("Docstring has multiple args with the same name! %s", arg_name)
        ds_param = ds_param[0]
        type_names, desc_names = _get_candidate_type_names_from_docstring(ds_param)
        docstring_type = _get_docstring_type(type_names)
        if not docstring_type:
            docstring_type = _get_docstring_type(desc_names)
        if docstring_type is not None:
            log.debug2("Found type from docstring for %s: %s", arg_name, docstring_type)
            return docstring_type
    else:
        log.warning("Found no parameter named %s:%s", arg_name, fn.__name__)
    return None


def _get_candidate_type_names_from_docstring(
    param: Optional[docstring_parser.common.DocstringParam],
) -> Tuple[List[str], List[str]]:
    if param is None:
        return [], []

    # Check the official 'type_name'
    candidate_type_names = []
    candidate_types_from_description = []
    if param.type_name is not None:
        candidate_type_names.append(param.type_name)

    # If not in type_name, try parsing our convention from the
    # description
    if param.description is not None:
        candidate_types_from_description.extend(
            [
                val
                for val in param.description.split("\n")[0].split()
                if val not in ["or", "|"]
            ]
        )
    log.debug3(
        "Candidate type names: %s, %s",
        candidate_type_names,
        candidate_types_from_description,
    )

    return candidate_type_names, candidate_types_from_description


def _get_docstring_type(
    candidate_type_names: List[str],
) -> Optional[Type]:
    """Given a parsed docstring parameter, look in all of the possible places
    for the actual type
    """

    log.debug2(
        "Candidate type names for docstring parsing are: %s", candidate_type_names
    )

    # If we can't find the name in either place, we're done
    if not candidate_type_names:
        log.debug2("Could not find type name from docstring")
        return None

    # Check all candidate type names
    valid_candidates = []
    for type_name in candidate_type_names:
        # Check for builtin types
        builtin_type = getattr(builtins, type_name, None)
        if builtin_type is not None:
            valid_candidates.append(builtin_type)
            log.debug2(f"Found valid candidate type: {builtin_type}")
            continue

        # Try to find things like "list(str)"
        # List[str]???
        nested_type = _extract_nested_type(type_name)
        if nested_type is not None:
            valid_candidates.append(nested_type)
            log.debug2(f"Found valid nested type: {nested_type}")
            continue

        # Try to spelunk down `sys.modules` for the type. This should work if it is fully qualified
        candidate_type = _extract_type_from_pymodule(sys.modules, type_name)
        if candidate_type is not None:
            valid_candidates.append(candidate_type)
            log.debug2(f"Found valid candidate type on sys.modules: {candidate_type}")
            continue

        # If the type was not fully qualified (like a `ProducerId`), look in a couple well known
        # places - the caikit core data model itself
        candidate_type = _extract_type_from_pymodule(caikit.core.data_model, type_name)
        if candidate_type is not None:
            valid_candidates.append(candidate_type)
            log.debug2(
                # pylint: disable=line-too-long
                f"Found valid candidate type on caikit.core.data_model: {candidate_type}"
            )
            continue
        # ...And the data model within the interfaces, including those defined in library
        try:
            candidate_type = DataBase.get_class_for_name(type_name)
        except ValueError:
            log.debug2(f"Data model match failed on {candidate_type}, continuing")
            continue
        if candidate_type is not None:
            valid_candidates.append(candidate_type)
            log.debug2(f"Found valid data model candidate type: {candidate_type}")
            continue

    log.debug3("valid candidates %s", valid_candidates)

    # If valid candidates were found, return either the single or a Union
    if valid_candidates:
        if len(valid_candidates) == 1:
            return valid_candidates[0]

        # pylint: disable=unnecessary-dunder-call
        return Union.__getitem__(tuple(valid_candidates))
    log.debug2(
        "Unable to pull type name [%s]",
        candidate_type_names,
    )


def _extract_nested_type(type_name: str) -> Optional[Type]:
    type_name = type_name.replace("[", "(").replace("]", ")")
    is_a_list = type_name.lower().startswith("list")
    if is_a_list:
        start_child_type_name = type_name.find("(") + 1
        end_child_type_name = type_name.rfind(")")
        child_type_name = type_name[start_child_type_name:end_child_type_name]

        child_type = _get_docstring_type(candidate_type_names=[child_type_name])
        if child_type:
            return List[child_type]


def _extract_type_from_pymodule(
    py_module: Union[Type, Dict], type_name: str
) -> Optional[Type]:
    """This walks down a type hierarchy to try to find the concrete type given an input string name

    Args:
        py_module (Type | Dict): A python module, or dictionary of modules, to
            start walking to find "type_name"
        type_name (str): The name of the type that we're trying to find. e.g.
            "caikit.core.data_model.ProducerId"

    Returns:
        Optional[Type]: The type of "type_name", or None if it cannot be found
    """
    output_type = py_module

    for part in type_name.split("."):
        if output_type != sys.modules:
            log.debug2(f"Looking for part {part} in {output_type}")
        else:
            log.debug2(f"Looking for part {part} in sys.modules")

        if isinstance(output_type, dict):
            output_type = output_type.get(part, None)
        else:
            output_type = getattr(output_type, part, None)
        if output_type is None:
            if not isinstance(py_module, dict):
                log.debug2(
                    "Couldn't find type name [%s] as an attribute on [%s]",
                    type_name,
                    py_module,
                )
            return None
    if output_type not in [None, py_module]:
        return output_type



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for producer priority"""


# Local
from caikit.core.data_model import ProducerId
from caikit.interfaces.common.data_model.producer import ProducerPriority

## Setup #####################################################################
PROD_ID_1 = ProducerId("foo", "1.2.3")
PROD_ID_2 = ProducerId("bar", "1.2.3")

## Tests #####################################################################


def test_proto_roundtrip():
    producer_priority = ProducerPriority(producers=[PROD_ID_1, PROD_ID_2])
    producer_priority_roundtrip = ProducerPriority.from_proto(
        producer_priority.to_proto()
    )
    assert producer_priority_roundtrip.producers[0] == PROD_ID_1
    assert producer_priority_roundtrip.producers[1] == PROD_ID_2


def test_json_roundtrip():
    producer_priority = ProducerPriority(producers=[PROD_ID_1, PROD_ID_2])
    producer_priority_roundtrip = ProducerPriority.from_json(
        producer_priority.to_json()
    )
    assert producer_priority_roundtrip.producers[0] == PROD_ID_1
    assert producer_priority_roundtrip.producers[1] == PROD_ID_2



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module holds the @module decorator which is used to decorate a
caikit.module
"""

# Standard
from typing import Dict, Optional, Type, Union
import collections

# Third Party
import semver

# First Party
import alog

# Local
from ..data_model import ProducerId
from ..exceptions import error_handler
from ..registries import module_backend_registry, module_backend_types, module_registry
from ..signature_parsing import CaikitMethodSignature
from ..task import TaskBase
from .base import ModuleBase
import caikit.core

log = alog.use_channel("MODULE_DEC")
error = error_handler.get(log)


# This is used by individual caikit.module implementations,
# to define what backend type models they can support at load time.
SUPPORTED_LOAD_BACKENDS_VAR_NAME = "SUPPORTED_LOAD_BACKENDS"


def module(
    id=None,  # pylint: disable=redefined-builtin
    name=None,
    version=None,
    task: Type[TaskBase] = None,
    backend_type="LOCAL",
    base_module: Union[str, Type[ModuleBase]] = None,
    backend_config_override: Optional[Dict] = None,
):
    f"""Apply this decorator to any class that should be treated as a caikit module
     (i.e., extends`{caikit.core.ModuleBase}) and registered with caikit.core so that the library
     "knows" the class is a caikit module and is capable of loading instances of the module.

    Args:
        id:  str
            A UUID to use when registering this module with caikit.core
            Not required if based on another caikit module using `base_module`
        name:  str
            A human-readable name for the module
            Not required if based on another caikit module using `base_module`
        version:  str
            A SemVer for the module
            Not required if based on another caikit module using `base_module`
        task:  Type[TaskBase]
            An ML task class that this module is an implementation for
            Not required if based on another caikit module using `base_module`
        backend_type: backend_type
            Associated backend type for the module.
            Default: `LOCAL`
        base_module: str | ModuleBase
            If this module is based on a different caikit module, provide name
            of the base module.
            Default: None
        backend_config_override: Dict
            Dictionary containing configuration required for the specific backend.
            Default: None

    Returns:
        A decorated version of the class to which it was applied, after registering the
        class as a valid module with caikit.core
    """
    base_module_class = None
    # Flag to store if the current module is a backend implementation
    # of an existing module or not
    backend_module_impl = False

    # No mutable default
    backend_config_override = backend_config_override or {}

    if any([id is None, version is None or name is None]):
        error.type_check(
            "<COR87944440E>",
            str,
            type(ModuleBase),
            allow_none=False,
            base_module=base_module,
        )
        error.type_check(
            "<COR60584425E>",
            dict,
            allow_none=True,
            backend_config_override=backend_config_override,
        )

        # If the base_module is a string, assume that it is the module_id of the
        # base module
        if isinstance(base_module, str):
            module_id = base_module
            error.value_check(
                "<COR09479833E>",
                module_id in module_registry(),
                "Unknown base module id: {}",
                module_id,
            )
            base_module_class = module_registry()[module_id]

        # If base_module is a type, validate that it derives from ModuleBase and
        # use its MODULE_ID
        elif isinstance(base_module, type):
            if not issubclass(base_module, ModuleBase):
                error(
                    "<COR20161747E>",
                    f"base_module [{base_module}] does not derive from ModuleBase",
                )

            base_module_class = base_module

        # TODO: Add support for inheritance of backend implementation
        # i.e if a module inherits from base_module

        id = base_module_class.MODULE_ID
        version = base_module_class.MODULE_VERSION
        name = base_module_class.MODULE_NAME
        task = base_module_class.TASK_CLASS
        backend_module_impl = True

    error.type_check("<COR54118928E>", str, id=id, name=name, version=version)
    error.subclass_check("<COR90789722E>", task, TaskBase, allow_none=True)

    semver.VersionInfo.parse(version)  # Make sure this is a valid SemVer

    def decorator(cls_):
        # Verify this is a valid module type (inherits from the wrapped base class)
        if not issubclass(cls_, caikit.core.ModuleBase):
            error(
                "<COR32401861E>",
                TypeError(
                    f"`{cls_.__name__}` does not extend ModuleBase",
                ),
            )

        # Add attributes to the implementation class
        cls_.MODULE_ID = id  # Module ID == Module Type ID
        cls_.MODULE_NAME = name  # Module Name == Module Type Name
        cls_.MODULE_VERSION = version  # Module Version == Module Type Version
        classname = f"{cls_.__module__}.{cls_.__qualname__}"
        cls_.MODULE_CLASS = classname
        cls_.PRODUCER_ID = ProducerId(cls_.MODULE_NAME, cls_.MODULE_VERSION)

        # Tasks: check to see if a super-class has one as well and that they match:
        tasks = {
            class_.TASK_CLASS for class_ in cls_.mro() if hasattr(class_, "TASK_CLASS")
        }
        if len(tasks) > 1:
            error(
                "<COR17197749E>",
                TypeError(
                    f"Class {cls_} has multiple task definitions in class hierarchy"
                ),
            )
        if tasks:
            parent_task = tasks.pop()
            if task and task != parent_task:
                error(
                    "<COR44943734E>",
                    TypeError(
                        f"Class {cls_} has task {task} but superclass has task "
                        f"{parent_task}"
                    ),
                )
            cls_.TASK_CLASS = parent_task
        else:
            cls_.TASK_CLASS = task

        # Parse the `train` and `run` signatures
        cls_.RUN_SIGNATURE = CaikitMethodSignature(cls_, "run")
        cls_.TRAIN_SIGNATURE = CaikitMethodSignature(cls_, "train")
        cls_._INFERENCE_SIGNATURES = []

        # If the module has a task, validate it:
        if cls_.TASK_CLASS:
            if not cls_.TASK_CLASS.has_inference_method_decorators(module_class=cls_):
                # Hackity hack hack - make sure at least one flavor is supported
                validated = False
                validation_errs = []
                for input_streaming, output_streaming in [
                    [False, False],
                    [True, True],
                    [False, True],
                ]:
                    try:
                        cls_.TASK_CLASS.validate_run_signature(
                            cls_.RUN_SIGNATURE, input_streaming, output_streaming
                        )
                        validated = True
                        cls_._INFERENCE_SIGNATURES.append(
                            (input_streaming, output_streaming, cls_.RUN_SIGNATURE)
                        )
                        break
                    except (ValueError, TypeError) as e:
                        validation_errs.append(e)
                if not validated:
                    raise validation_errs[0]

            cls_.TASK_CLASS.deferred_method_decoration(cls_)

        # If no backend support described in the class, add current backend
        # as the only backend that can load models trained by this module
        cls_.SUPPORTED_LOAD_BACKENDS = getattr(
            cls_, SUPPORTED_LOAD_BACKENDS_VAR_NAME, [backend_type]
        )

        # Set its own backend_type as an attribute
        setattr(cls_, "BACKEND_TYPE", backend_type)

        # Verify UUID and add this module to the module registry
        if not backend_module_impl:
            if cls_.MODULE_ID in module_registry():
                error(
                    "<COR30607646E>",
                    RuntimeError(
                        "MODULE_ID `{}` conflicts for classes `{}` and `{}`".format(
                            cls_.MODULE_ID,
                            cls_.__name__,
                            module_registry()[cls_.MODULE_ID].__name__,
                        )
                    ),
                )
            module_registry()[cls_.MODULE_ID] = cls_

        # Register backend
        _register_module_implementation(
            cls_, backend_type, cls_.MODULE_ID, backend_config_override
        )

        return cls_

    return decorator


## Implementation Details ######################################################


def _register_module_implementation(
    implementation_class: type,
    backend_type: str,
    module_id: str,
    backend_config_override: Dict = None,
):
    """This function will register the mapping for the given module_id and
    backend_type to the implementation class

    Args:
        implementation_class (type): The class that is used to implement this
            backend type for the given module_id
        backend_type (str): Value from MODULE_BACKEND_TYPES that indicates the
            backend that this class implements
        module_id (str): The module_id from the caikit.core module registry that
            this class overloads
        backend_config_override (Dict): Dictionary containing essential
            overrides for the backend config. This will get stored with the
            implementation_class class name and will automatically get picked up
            and merged with other such configs for a specific backend

    """

    backend_config_override = backend_config_override or {}

    log.debug(
        "Registering backend [%s] implementation for module [%s]",
        backend_type,
        module_id,
    )

    error.value_check(
        "<COR86780140E>",
        backend_type in module_backend_types(),
        "Cannot override implementation of {} for unknown backend type {}",
        module_id,
        backend_type,
    )

    core_class = module_registry().get(module_id)
    if core_class is None:
        # TODO! Inject a dummy entry that will raise on usage
        # Info level log: this is normal behavior if backend module is imported
        # before the base module
        log.info(
            "<COR86369940I>",
            "No core class found for new backend %s with module ID %s",
            implementation_class.__name__,
            module_id,
        )

    # Do the registration!
    module_type_mapping = module_backend_registry().setdefault(module_id, {})

    # Make sure this is not an overwrite of an existing registration
    existing_type = module_type_mapping.get(backend_type)
    assert (
        existing_type is None or existing_type is implementation_class
    ), f"Registration conflict! ({module_id}, {backend_type}) already registered as {existing_type}"

    BackendConfig = collections.namedtuple(
        "BackendConfig", "impl_class backend_config_override"
    )

    module_type_mapping[backend_type] = BackendConfig(
        impl_class=implementation_class, backend_config_override=backend_config_override
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data structures for classification representations"""

# Standard
from typing import List, Optional

# Third Party
import numpy as np

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber
import alog

# Local
from ....core import DataObjectBase, dataobject
from .package import NLP_PACKAGE
from .text_generation import FinishReason

log = alog.use_channel("DATAM")


@dataobject(package=NLP_PACKAGE)
class ClassificationTrainRecord(DataObjectBase):
    """A classification training record consisting of a single train instance."""

    text: Annotated[str, FieldNumber(1)]  # Text to be classified
    labels: Annotated[
        List[str], FieldNumber(2)
    ]  # Class labels to be learnt for the text


@dataobject(package=NLP_PACKAGE)
class ClassificationResult(DataObjectBase):
    """A single classification prediction."""

    label: Annotated[str, FieldNumber(1)]  # Predicted relevant class name
    score: Annotated[
        float, FieldNumber(2)
    ]  # The confidence-like score of this prediction in [0, 1]


@dataobject(package=NLP_PACKAGE)
class ClassificationResults(DataObjectBase):
    """Classification results generated from a text and consisting multiple classes."""

    results: Annotated[
        List[ClassificationResult], FieldNumber(1)
    ]  # List of classifications for a text


# NOTE: This is meant to align with the HuggingFace token classification task:
# https://huggingface.co/docs/transformers/tasks/token_classification#inference
# The field `word` does not necessarily correspond to a single "word",
# and `entity` may not always be applicable beyond "entity" in the NER
# (named entity recognition) sense
@dataobject(package=NLP_PACKAGE)
class TokenClassificationResult(DataObjectBase):
    """A single token classification prediction."""

    start: Annotated[int, FieldNumber(1)]  # Beginning offset of the token
    end: Annotated[int, FieldNumber(2)]  # Ending offset of the token
    word: Annotated[str, FieldNumber(3)]  # Text referenced by this token
    entity: Annotated[
        str, FieldNumber(4)
    ]  # Predicted relevant class name for the token
    entity_group: Annotated[str, FieldNumber(5)]  # Aggregate label, if applicable
    score: Annotated[
        float, FieldNumber(6)
    ]  # The confidence-like score of this classification prediction in [0, 1]
    token_count: Annotated[
        Optional[int], FieldNumber(7)
    ]  # Length of tokens in the text


@dataobject(package=NLP_PACKAGE)
class TokenClassificationResults(DataObjectBase):
    """Token classification results generated from a text and consisting multiple classes."""

    results: Annotated[List[TokenClassificationResult], FieldNumber(1)]


@dataobject(package=NLP_PACKAGE)
class TokenClassificationStreamResult(TokenClassificationResults):
    """
    Streaming token classification results that indicates up to where in stream is processed.
    """

    processed_index: Annotated[
        int, FieldNumber(2)
    ]  # Result index up to which text is processed


@dataobject(package=NLP_PACKAGE)
class ClassifiedGeneratedTextResult(DataObjectBase):
    """Classification result on text produced by a text generation model, contains
    information from the original text generation output as well as the result of
    classification on the generated text.
    """

    @dataobject(package=NLP_PACKAGE)
    class TextGenTokenClassificationResults(DataObjectBase):
        input: Annotated[Optional[List[TokenClassificationResult]], FieldNumber(10)]
        output: Annotated[Optional[List[TokenClassificationResult]], FieldNumber(20)]

    generated_text: Annotated[Optional[str], FieldNumber(1)]  # The generated text
    token_classification_results: Annotated[
        Optional[TextGenTokenClassificationResults], FieldNumber(2)
    ]  # Token classification results for input and generated text
    finish_reason: Annotated[
        Optional[FinishReason], FieldNumber(3)
    ]  # Reason as to why text generation stopped
    generated_token_count: Annotated[
        Optional[int], FieldNumber(4)
    ]  # Length of generated tokens sequence
    seed: Annotated[
        Optional[np.uint64], FieldNumber(5)
    ]  # The random seed used for text generation
    input_token_count: Annotated[Optional[int], FieldNumber(6)]


@dataobject(package=NLP_PACKAGE)
class ClassifiedGeneratedTextStreamResult(ClassifiedGeneratedTextResult):
    """
    Streaming classification on generated text result that indicates up to where in stream
    is processed.
    """

    processed_index: Annotated[
        Optional[int], FieldNumber(7)
    ]  # Result index up to which text is processed



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared functionality and interfaces used by *all* moduless.
"""

# TODO: Review to see if the lines in this file can be reduced or
# broken out into another file (e.g. Python sub-modules)
# pylint: disable=too-many-lines

# abstract methods may define arguments but not use them
# pylint: disable=unused-argument

# Standard
from io import BytesIO
from typing import Any, Dict, List, Optional, Union
import collections
import os
import shutil
import tempfile
import time
import types

# First Party
import alog

# Local
from ..data_model import DataBase, DataStream
from ..exceptions import error_handler
from ..exceptions.validation_error import DataValidationError
from ..toolkit import fileio
from .config import ModuleConfig
from .loader import ModuleLoader
from .meta import _ModuleBaseMeta
from caikit import core

log = alog.use_channel("MODULE")
error = error_handler.get(log)


class ModuleBase(metaclass=_ModuleBaseMeta):
    """Abstract base class from which all modules should inherit."""

    def __init__(self):
        """Construct a new model."""
        # Set up an empty metadata dictionary, to be:
        # - populated with metadata from `config.yml` files on `load`, and
        # - saved back to `config.yml` files on `save`
        self._metadata = {}
        # Keep an indicator of the backend used to load this model
        self._load_backend = None

    #############
    ## Utility ##
    #############

    @property
    def metadata(self) -> Dict[str, Any]:
        """This module's metadata.

        Returns:
            Dict[str, Any]: A dictionary of this module's metadata

            TODO: Can this be a `ModuleConfig` object instead? (or aconfig.Config)?
        """
        if not hasattr(self, "_metadata") or self._metadata is None:
            self._metadata = {}
        return self._metadata

    def set_load_backend(self, load_backend):
        """Method used by the model manager to indicate the load backend that
        was used to load this module
        """
        self._load_backend = load_backend

    @classmethod
    def get_inference_signature(
        cls, input_streaming: bool, output_streaming: bool
    ) -> Optional["caikit.core.signature_parsing.CaikitMethodSignature"]:
        """Returns the inference method signature that is capable of running the module's task
        for the given flavors of input and output streaming
        """
        for in_streaming, out_streaming, signature in cls._INFERENCE_SIGNATURES:
            if in_streaming == input_streaming and out_streaming == output_streaming:
                return signature
        return None

    @property
    def load_backend(self):
        """Get the backend instance used to load this module. This can be used
        in module implementations that require use of a specific backend at
        inference time.
        """
        return self._load_backend

    ###################
    ## Instantiation ##
    ###################

    @classmethod
    @alog.logged_function(log.debug)
    def bootstrap(cls, *args, **kwargs):
        """Bootstrap a module. This method can be used to initialize the module
        from artifacts created outside of a particular caikit library
        """
        error(
            "<COR92634438E>",
            NotImplementedError("This is not available in this module."),
        )

    @classmethod
    @alog.logged_function(log.debug)
    def load(
        cls,
        model_path: Union[str, ModuleConfig],
        *args,
        **kwargs,
    ) -> "ModuleBase":
        """Load a new instance of workflow from a given model_path

        Args:
            model_path (Union[str, ModuleConfig]): Path to saved model or
                in-memory ModuleConfig
        Returns:
            model (ModuleBase): A new instance of this module class
        """
        return cls._load(ModuleLoader(model_path), *args, **kwargs)

    @classmethod
    def _load(cls, module_loader, *args, **kwargs):
        """Load a model."""
        error(
            "<COR88356566E>",
            NotImplementedError("This is not available in this module."),
        )

    @classmethod
    def timed_load(cls, *args, **kwargs):
        """Time a model `load` call.

        Args:
            *args (list): Will be passed to `self.load`.
            **kwargs (dict): Will be passed to `self.load` -- the only way to
                pass arbitrary arguments to `self.load` from this function.

        Returns:
            int, caikit.core._ModuleBase: The first return value is the total
                time spent in the `self.load` call. The second return value is
                the loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
        # get initial values
        start_time = time.time()
        # We are calling caikit.core over cls.load because we need to figure out
        # what instance the model belongs to
        model = core.load(*args, **kwargs)
        time_passed = time.time() - start_time
        return time_passed, model

    def validate_loaded_model(self, *args):
        """Validate a loaded model."""
        error(
            "<COR56275627E>",
            NotImplementedError("This is not available in this module."),
        )

    ###################
    ## Serialization ##
    ###################

    @alog.logged_function(log.debug)
    def save(self, model_path: str, *args, **kwargs):
        """Save a model.

        Args:
            model_path (str): Path on disk to export the model to.
        """
        error(
            "<COR58632237E>",
            NotImplementedError("This is not available in this module."),
        )

    @alog.logged_function(log.debug)
    def as_file_like_object(self, *args, **kwargs) -> BytesIO:
        """Produces a file-like object corresponding to a zip archive affiliated with a given
        model. This method wraps is functionally similar to .save() - it saves a model into
        a temporary directory and produces a zip archive, then loads the result as a io.BytesIO
        object. The result of this function is also compatible with .load(), and cleanup is
        handled automatically.

        Args:
            *args, **kwargs (dict): Optional keyword arguments for saving.
        Returns:
            io.BytesIO: File like object holding an exported model in memory as
                a io.BytesIO object.
        """
        return BytesIO(self.as_bytes(*args, **kwargs))

    @alog.logged_function(log.debug)
    def as_bytes(self, *args, **kwargs) -> bytes:
        """Produces a bytes object corresponding to a zip archive affiliated with a given
        model. This method wraps is functionally similar to .save() - it saves a model into
        a temporary directory and produces a zip archive, then loads the result as a bytes
        object. The result of this function is also compatible with .load(), and cleanup is
        handled automatically.

        Args:
            *args, **kwargs (dict): Optional keyword arguments for saving.
        Returns:
            bytes: bytes object holding an exported model in memory.
        """
        # Open a temporary directory & do all operations relative to that temporary directory.
        with tempfile.TemporaryDirectory() as ephemeral_model_path:
            # Save the model to the temporary directory
            model_path = os.path.join(ephemeral_model_path, ".model")
            zip_path = os.path.join(ephemeral_model_path, ".archive")
            zip_path_with_ext = zip_path + ".zip"
            self.save(model_path, *args, **kwargs)
            try:
                # Compress the model to a zip archive in the temporary directory
                shutil.make_archive(zip_path, "zip", model_path)
                # Load the zip archive bytes into memory as a file-like object and clean up any disk
                # objects (NOTE: it is safe to delete the archive once we extract the bytes).
                with open(zip_path_with_ext, "rb") as handle:
                    in_memory_archive = handle.read()
            except PermissionError:
                error(
                    "<COR80051233E>",
                    PermissionError(
                        "Unable to create archive to be loaded into memory."
                    ),
                )
            return in_memory_archive

    ###############
    ## Inference ##
    ###############

    @alog.logged_function(log.debug)
    def run(self, *args, **kwargs):
        """Run a model - this typically makes a single prediction and returns an object from the
        data model.
        """
        error(
            "<COR80640037E>",
            NotImplementedError("This is not available in this module."),
        )

    @alog.logged_function(log.debug)
    def run_batch(self, *args, **kwargs):
        """Run a model in batch mode - this typically ingests an iterable of inputs that can be
        applied to run & returns a list of data model objects that run ordinarily returns. A module
        may override this method to provide faster evaluation capabilities, e.g., by leveraging
        vectorization during prediction.

        All provided args and kwargs that should be expanded with the batch should be provided as
        prebatched iterables. If a provided arg/kwarg is not provided as an iterable, it will be
        passed as is to all self contained run calls, which may be the case in some rare cases,
        such as runtime explanability enablement.

        This function is intentionally kept as simple as possible. In order to maintain its
        simplicity, all argument iterables must be the same length, where the length of every
        provided iterable is presumed to be the batch size. If an iterable must be passed as
        arg to each run call, batch run must be called by wrapping it in another iterable and
        duplicating the iterable arg to match the size, or (ideally) overridden in the subclass
        as necessary.

        Args:
            *args: Variable length argument list to be passed directly to run().
            **kwargs: Arbitrary keyword arguments to be passed directly to run().
        Returns:
            tuple: Iterable of prediction outputs, run as a batch.
        """
        predictions = []
        fixed_args = {}
        expanded_args = {}
        fixed_kwargs = {}
        expanded_kwargs = {}
        batch_size = self._validate_and_extract_batch_size(*args, **kwargs)
        # Keep map args to indices - mark iterable nonstrings as expandable args
        for idx, arg in enumerate(args):
            if self._is_expandable_iterable(arg):
                expanded_args[idx] = arg
            else:
                fixed_args[idx] = arg
        # And do the same with kwargs
        for kwarg_key, kwarg_value in kwargs.items():
            if self._is_expandable_iterable(kwarg_value):
                expanded_kwargs[kwarg_key] = kwarg_value
            else:
                fixed_kwargs[kwarg_key] = kwarg_value

        for idx in range(batch_size):
            run_args = self._build_args_for_default_run_with_batch(
                fixed_args, expanded_args, idx
            )
            run_kwargs = self._build_kwargs_for_default_run_with_batch(
                fixed_kwargs, expanded_kwargs, idx
            )
            run_out = self.run(*run_args, **run_kwargs)
            predictions.append(run_out)
        return tuple(predictions)

    def timed_run(self, *args, num_seconds=None, num_iterations=None, **kwargs):
        """Time a number of runs over set seconds or iterations.

        Args:
            *args (list): Will be passed to `self.run`.
            num_seconds (int): Minimum number of seconds to run timed_run over.
                Will most likely be more than this value due to its waiting for
                the each call to `self.run` to finish.
            num_iterations (int): Minimum number of iterations to run timed_run
                over. Will run exactly this many times.
            **kwargs (dict): Will be passed to `self.run`.

        Returns:
            int, int, caikit.core.data_model.DataBase: The first return value is
                the total time spent in the `self.run` loop. The second return
                value is the total number of calls to `self.run` were made. The
                return value is the output of the module's run method

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_run("some example text", num_seconds=60)`

        By default it will run for greater than or equal to 120 seconds.
        """
        # default to running for 120 seconds
        if not (num_seconds or num_iterations):
            num_seconds = 120

        # get initial values
        start_time = time.time()
        iterations_passed = 0
        time_passed = time.time() - start_time

        # stop on seconds or iterations depending on input arguments
        # pylint: disable=unnecessary-lambda-assignment
        continue_condition = (
            lambda t_p, i_p: t_p <= num_seconds if num_seconds else i_p < num_iterations
        )
        response = None

        while continue_condition(time_passed, iterations_passed):
            # use model's run method
            response = self.run(*args, **kwargs)

            # increment output values
            time_passed = time.time() - start_time
            iterations_passed += 1

        return time_passed, iterations_passed, response

    def stream(self, data_stream, *args, **kwargs):
        """Lazily evaluate a run() on a given model by constructing a new data stream generator
        from the results. Note that we do not allow datastreams in args/kwargs. In rare cases,
        this may mean that stream() is not available, e.g., for keywords extraction. In these
        cases, stream() should be overridden in the subclass (module implementation) to allow
        and expand along multiple data streams.

        Args:
            data_stream (caikit.core.data_model.DataStream): Datastream to be
                lazily sequentially processed by the module under consideration.
            *args: Variable length argument list to be passed directly to run().
            **kwargs: Arbitrary keyword arguments to be passed directly to run().
        Returns:
            protobufs: A DataBase object.
        """
        error.type_check("<COR98214589E>", DataStream, data_stream=data_stream)
        # Ensure that no args/kwargs are DataStreams, since these get passed to stream()
        run_argvals = list(args) + list(kwargs.values())
        if any(isinstance(arg, DataStream) for arg in run_argvals):
            error(
                "<COR28828273E>",
                ValueError(
                    "Only one DataStream may be passed when invoking module stream()"
                ),
            )
        # TODO: Add .run_batch() integration
        return DataStream(
            lambda: (self.run(data_item, *args, **kwargs) for data_item in data_stream)
        )

    ##############
    ## Training ##
    ##############

    @classmethod
    @alog.logged_function(log.debug)
    def train(cls, *args, **kwargs):
        """Train a model."""
        error(
            "<COR44977721E>",
            NotImplementedError("This is not available in this module."),
        )

    @classmethod
    def validate_training_data(
        cls, training_data: Union[str, DataStream], limit: int = -1
    ) -> List[DataValidationError]:
        """Validate a set of training data, passed as a filename or as a data stream.
        Return up to `limit` number of DataValidationErrors
        """
        error(
            "<COR56285627E>",
            NotImplementedError("This is not available in this module."),
        )

    ################
    ## Evaluation ##
    ################

    # will be used to evaluate modules; defined in sub-classes
    evaluation_type = None
    # first arg is "self", unfortunately; TODO: get rid of that somehow
    evaluator = None

    @staticmethod
    def find_label_func(*_args, **_kwargs):
        """Function used to extract "label" from a prediction/result of a module's .run method.
        Define if you wish to have more specific evaluation metrics. Implemented in subclass.
        """
        raise NotImplementedError("Func not implemented")

    @staticmethod
    def find_label_data_func(*_args, **_kwargs):
        """Function used to extract data belonging to class "label" from a prediction/result
        of a module's .run method. Define if you wish to have more specific evaluation metrics.
        Implemented in subclass.
        """
        raise NotImplementedError("Func not implemented")

    def evaluate_quality(
        self,
        dataset_path,
        preprocess_func=None,
        detailed_metrics=False,
        labels=None,
        partial_match_metrics=False,
        max_hierarchy_levels=3,
        **kwargs,
    ):
        """Run quality evaluation for instance of module

        Args:
            dataset_path (str): Path to where the input "gold set" dataset
                lives. Most often this is .json file.
            preprocess_func (method): Function used as proxy for any preliminary
                steps that need to be taken to run the model on the input text.
                This helper function ultimately leads to the input to this
                module and may involve executing other modules.
            detailed_metrics: boolean (Optional, defaults to False)
                Only for 'keywords'. Include partial scores and scores over every text in document.
            labels: list (Optional, defaults to None)
                Optional list of class labels to evaluate quality on. By default evaluation is done
                over all class labels. Using this, you can explicitly mention only a subset of
                labels to include in the quality evaluation.
            partial_match_metrics: boolean (Optional, defaults to False)
                Include partial match micro avg F1.
            max_hierarchy_levels (int): Used in hierarchical multilabel
                multiclass evaluation only. The number of levels in the
                hierarchy to run model evaluation on, in addition to complete
                matches.
            *args, **kwargs: Optional arguments which can be used by goldset/prediction
                set extraction.
                Nonekeyword arguments: `block_level`: str
                    For any module that has pre processing steps in the
                    middle of raw text and actual module input, use the input from gold standard
                    labels instead of a pre-process function. Useful for measuring quality for the
                    'block' alone (instead of the module + pre_process pipeline)
        Returns:
            dict: Dictionary of results provided by the `self.evaluator.run`
                function, depending on the associated `evaluation_type`. Reports
                things like precision, recall, and f1.
        """
        # 1) load dataset
        dataset = self._load_evaluation_dataset(dataset_path)

        # 2) verify dataset
        error.type_check("<COR14030040E>", collections.abc.Iterable, dataset=dataset)

        # 3) extract gold set predictions
        # pylint: disable=assignment-from-no-return
        gold_set = self._extract_gold_set(dataset)
        gold_annos = self._extract_gold_annotations(gold_set)

        # 4) obtain pred set predictions
        # pylint: disable=assignment-from-no-return
        pred_set = self._extract_pred_set(
            dataset, preprocess_func=preprocess_func, **kwargs
        )
        pred_annos = self._extract_pred_annotations(pred_set)

        # 5) initialize evaluator
        # pylint: disable=not-callable
        evaluator = self.evaluator(gold_annos, pred_annos)

        # 6) run evaluator
        results = evaluator.run(
            self.evaluation_type,
            self.find_label_func,
            self.find_label_data_func,
            detailed_metrics,
            labels,
            partial_match_metrics,
            max_hierarchy_levels,
        )

        # 7) generate report
        report = self._generate_report(results, gold_set)

        # 8) return report
        return report

    ## Implementation Details ##################################################

    @staticmethod
    def _is_expandable_iterable(arg):
        """Check to see if something is a list / tuple of data model objects or strings. If it is,
        we consider it "expandable", meaning that one element of the iterable to one run call. In
        contrast, if something is not expandable, it will be passed as is to each call.

        Args:
            arg (any): Argument to run_batch being considered.
        Returns:
            bool: True if the argument is a compatible iterable, False
                otherwise.
        """
        # Throw if generators are passed - can't imagine any situation (for now) where this is
        # something that someone is doing on purpose, so we are a bit specific about this error.
        if isinstance(arg, types.GeneratorType):
            error(
                "<COR28071103E>",
                ValueError("Generator types are incompatible with .run_batch"),
            )
        if isinstance(arg, DataStream):
            error(
                "<COR75305604E>",
                ValueError("Data streams are incompatible with .run_batch"),
            )
        if isinstance(arg, (tuple, list)):
            return all(isinstance(obj, (str, DataBase)) for obj in arg)
        return False

    def _validate_and_extract_batch_size(self, *args, **kwargs):
        """Check to ensure that there's at least one iterable whose length is well defined,
        i.e., no generators, and that if multiple iterable arg/kwarg values are provided,
        they are all the same length.

        Args:
            *args: Variable length argument list to be passed directly to run().
            **kwargs: Arbitrary keyword arguments to be passed directly to run().
        Returns:
            int: Inferred batch size based on expandable iterables.
        """
        batch_size = None
        for _, arg in enumerate(args):
            batch_size = self._validate_arg_and_verify_batch_size(arg, batch_size)
        for _, arg_value in kwargs.items():
            batch_size = self._validate_arg_and_verify_batch_size(arg_value, batch_size)
        if batch_size is None:
            error("<COR36507545E>", ValueError("No iterable passed to batch predict"))
        return batch_size

    def _validate_arg_and_verify_batch_size(self, val, current_batch_size):
        """Check an arg value from args/kwargs. If we find that it's an expandable iterable, see
        if it conflicts with what we know about the inferred batch size so far.

        args:
            val (any): Argument / keyword argument value being inspected.
            current_batch_size (None | int): Current inferred batch size from
                previous args/kwargs, or None if no inferences have been made on
                other expandable iterables yet.
        Returns:
            None | inferred batch size.
        """
        if self._is_expandable_iterable(val):
            iter_batch_size = len(val)
            # Set the batch size if it's not set already. Raise if we have conflicting iterator
            # sizes. This will happen if the arg of a module run call has an iterable value. In
            # such cases, the subclass should override run_batch.
            if current_batch_size is None:
                return iter_batch_size
            if current_batch_size != iter_batch_size:
                err_str = "Iterables of varying length may not be passed to default batch predict."
                error("<COR98338604E>", ValueError(err_str))
        if current_batch_size:
            return current_batch_size

    @staticmethod
    def _build_args_for_default_run_with_batch(fixed_args, expanded_args, idx):
        """Build the non keyword arguments for run_batch's default implementation by expanding
        iterable args where possible, and grouping them with repeated noniterable arguments. The
        index correspondes to the current document under consideration.

        Args:
            fixed_args (dict): Noniterable args - common across all documents.
            expanded_args (dict): Iterable args - we'll need to index into this
                to get our doc arg.
            idx (int): Index of the document being considered.
        Returns:
            list: Args to be run for document [idx].
        """
        constructed_args = []
        if not expanded_args and not fixed_args:
            return constructed_args
        # Keys in arg dicts are positional arg number - get the biggest number arg we have
        max_exarg_idx = 0 if not expanded_args.keys() else max(expanded_args.keys())
        max_fixarg_idx = 0 if not fixed_args.keys() else max(fixed_args.keys())
        arg_count = max(max_exarg_idx, max_fixarg_idx)
        for arg_no in range(arg_count + 1):
            if arg_no in fixed_args:
                constructed_args.append(fixed_args[arg_no])
            elif arg_no in expanded_args:
                try:
                    constructed_args.append(expanded_args[arg_no][idx])
                except IndexError:
                    error(
                        "<COR94219537E>",
                        ValueError("idx {} exceeds extracted batch size".format(idx)),
                    )
            else:
                error(
                    "<COR68021708E>",
                    ValueError(
                        "arg number [{}] is missing from run_batch call".format(arg_no)
                    ),
                )
        return constructed_args

    @staticmethod
    def _build_kwargs_for_default_run_with_batch(fixed_kwargs, expanded_kwargs, idx):
        """Similar to the previous function, but for kwargs. Note that we can just clone our fixed
        kwargs instead of cycling through them, because order doesn't matter here.

        Args:
            fixed_args (dict): Noniterable valued kwargs - common across all
                documents.
            expanded_args (dict): Iterable valued kwargs - we'll need to index
                into these to get our doc kwarg.
        Returns:
            dict: Kwargs to be run for document [idx].
        """
        constructed_kwargs = fixed_kwargs.copy()
        for arg_name, iterable_arg_val in expanded_kwargs.items():
            try:
                constructed_kwargs[arg_name] = iterable_arg_val[idx]
            except IndexError:
                error(
                    "<COR51120124E>",
                    ValueError("idx {} exceeds extracted batch size".format(idx)),
                )
        return constructed_kwargs

    def _extract_gold_set(self, dataset):
        """Method for extracting gold set from dataset. Implemented in subclass.

        Args:
            dataset (object): In-memory version of whatever is loaded from on-
                disk. May be json, txt, etc.

        Returns:
            list: List of labels in the format of the module_type that is being
                called.
        """
        error(
            "<COR01455940E>",
            NotImplementedError("This is not available in this module."),
        )

    def _extract_pred_set(self, dataset, preprocess_func=None, **kwargs):
        """Method for extracting pred set from dataset. Implemented in subclass.

        Args:
            dataset (object): In-memory version of whatever is loaded from on-
                disk. May be json, txt, etc.
            preprocess_func (method): Function used as proxy for any preliminary
                steps that need to be taken to run the model on the input text.
                This helper function ultimately leads to the input to this
                module and may involve executing other modules.
            **kwargs (dict): Optional keyword arguments for prediction set extraction.
        Returns:
            list: List of labels in the format of the module_type that is being
                called.
        """
        error(
            "<COR95693719E>",
            NotImplementedError("This is not available in this module."),
        )

    @staticmethod
    def _load_evaluation_dataset(dataset_path):
        """Helper specifically for dataset loading.

        Args:
            dataset_path (str): Path to where the input 'gold set' dataset
                lives. Most often this is .json file.

        Returns:
            object: list, dict, or other python object, depending on the input
                dataset_path extension. Currently only supports `.json` and uses
                fileio from toolkit.
        """
        error.type_check("<COR33285197E>", str, dataset_path=dataset_path)

        if dataset_path.endswith(".json"):
            return fileio.load_json(dataset_path)

        # if all else fails
        error(
            "<COR81451234E>",
            ValueError("Unsure of how to load: {0}".format(dataset_path)),
        )

    @staticmethod
    def _extract_gold_annotations(gold_set):
        """Extract the core list of annotations that is needed for quality evaluation

        Args:
            gold_set (list)
        Returns:
            gold_annotations: list
        """
        return gold_set

    @staticmethod
    def _extract_pred_annotations(pred_set):
        """Extract the core list of predictions that is needed for quality evaluation

        Args:
            pred_set (list)
        Returns:
            pred_annotations: list
        """
        return pred_set

    @staticmethod
    def _generate_report(report, gold_set):
        """Generate the quality report output
        Args:
            report (dict)
            gold_set (list(dict))
        """
        return report



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.core.toolkit import F1Metrics, F1MetricsContainer, QualityEvaluator

# Unit Test Infrastructure
from tests.base import TestCaseBase
import caikit.core


class TestQualityEvaluation(TestCaseBase):
    def test___init__(self):
        q_er = caikit.core.toolkit.QualityEvaluator("abc", "def")
        self.assertEqual(q_er.gold, "abc")
        self.assertEqual(q_er.pred, "def")

    def test_multilabel_multiclass_evaluation(self):
        gold = [["a", "b", "c"], ["a"]]
        pred = [["b", "c"], ["b"]]

        def find_label_func(ele):
            return ele

        def find_label_data_func(data, label):
            return [ele for ele in data if ele == label]

        evaluation_type = caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        output = evaluator.run(evaluation_type, find_label_func, find_label_data_func)

        self.assertEqual(output["per_class_confusion_matrix"]["b"]["precision"], 0.5)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["recall"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["precision"], 0)
        self.assertEqual(round(output["macro_precision"], 3), 0.5)
        self.assertEqual(round(output["macro_recall"], 3), 0.667)
        self.assertEqual(round(output["macro_f1"], 3), 0.556)
        self.assertEqual(round(output["micro_precision"], 3), 0.667)
        self.assertEqual(round(output["micro_recall"], 3), 0.5)
        self.assertEqual(round(output["micro_f1"], 3), 0.571)

    def test_hierarchical_multilabel_multiclass_evaluation_no_hierarchy(self):
        """Test that hierarchical eval is consistent with normal eval with no label hierarchy"""
        gold = [["a", "b", "c"], ["a"]]
        pred = [["b", "c"], ["b"]]

        def label_func_builder(level):
            def find_label_func(ele):
                return ele

            return find_label_func

        def label_data_func_builder(level):
            def find_label_data_func(data, label):
                return [ele for ele in data if ele == label]

            return find_label_data_func

        evaluation_type = (
            caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS_HIERARCHICAL
        )
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        full_output = evaluator.run(
            evaluation_type, label_func_builder, label_data_func_builder
        )
        for output in full_output.values():
            self.assertEqual(
                output["per_class_confusion_matrix"]["b"]["precision"], 0.5
            )
            self.assertEqual(output["per_class_confusion_matrix"]["b"]["recall"], 1)
            self.assertEqual(output["per_class_confusion_matrix"]["a"]["precision"], 0)
            self.assertEqual(round(output["macro_precision"], 3), 0.5)
            self.assertEqual(round(output["macro_recall"], 3), 0.667)
            self.assertEqual(round(output["macro_f1"], 3), 0.556)
            self.assertEqual(round(output["micro_precision"], 3), 0.667)
            self.assertEqual(round(output["micro_recall"], 3), 0.5)
            self.assertEqual(round(output["micro_f1"], 3), 0.571)

    def test_hierarchical_multilabel_multiclass_evaluation_with_hierarchy(self):
        """Test that hierarchical eval is consistent with normal eval with a label hierarchy"""
        gold = [
            [("A sentence about basketball players", "/sports/basketball/players")],
            [("A sentence about basketball", "/sports/basketball")],
            [("A sentence about sports", "/sports")],
        ]
        pred = [
            [("A sentence about basketball players", "/sports/basketball")],
            [("A sentence about basketball", "/sports/baseball")],
            [("A sentence about sports", "/sports")],
        ]
        expected_metrics = {
            # Matching up to L1 - Matches everything
            "level_1": {"overall_tp": 3, "overall_fp": 0, "overall_fn": 0},
            # Matching up to L2 - Matches first and last example, misses middle one
            "level_2": {"overall_tp": 2, "overall_fp": 1, "overall_fn": 1},
            # Full matching - only one total match
            "level_3": {"overall_tp": 1, "overall_fp": 2, "overall_fn": 2},
            "level_all": {"overall_tp": 1, "overall_fp": 2, "overall_fn": 2},
        }

        # Helper for slicing up to subhierarchy
        def slice_up_to_level(level, label):
            if level is None:
                return label
            return "/".join(label.split("/")[: level + 1])

        # Define the hierarchical builder functions
        def label_func_builder(level):
            def find_label_func(ele):
                return slice_up_to_level(level, ele[1])

            return find_label_func

        def label_data_func_builder(level):
            def find_label_data_func(data, gold_label):
                def matches_up_to_level(label1, label2):
                    return slice_up_to_level(level, label1) == slice_up_to_level(
                        level, label2
                    )

                text, pred_label = data[0]
                if matches_up_to_level(gold_label, pred_label):
                    return [text]
                return []

            return find_label_data_func

        # Run the evaluator
        evaluation_type = (
            caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS_HIERARCHICAL
        )
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        full_output = evaluator.run(
            evaluation_type,
            label_func_builder,
            label_data_func_builder,
            max_hierarchy_levels=3,
        )

        full_output_2_levels = evaluator.run(
            evaluation_type,
            label_func_builder,
            label_data_func_builder,
            max_hierarchy_levels=2,
        )

        # Full output has N levels + a "level_all"
        self.assertEqual(len(full_output), 4)  # 3 levels + level_all
        self.assertEqual(len(full_output_2_levels), 3)  # 2 levels + level_all

        # For each hierarchy level, ensure that generated metrics match
        for level_key, actual_level_metrics in full_output.items():
            exp_level_metrics = expected_metrics[level_key]
            for metric_key in exp_level_metrics:
                self.assertEqual(
                    exp_level_metrics[metric_key], actual_level_metrics[metric_key]
                )

    def test_singlelabel_multiclass_evaluation_no_labels(self):
        gold = ["a", "b", "a", "b"]
        pred = ["b", "b", "b", "b"]
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        output = evaluator.run(
            evaluation_type=caikit.core.toolkit.EvalTypes.SINGLELABEL_MULTICLASS
        )

        self.assertEqual(output["micro_f1"], 0.5)
        self.assertEqual(output["micro_precision"], 0.5)
        self.assertEqual(output["micro_recall"], 0.5)
        self.assertEqual(output["macro_precision"], 0.25)
        self.assertEqual(output["macro_recall"], 0.5)
        self.assertEqual(round(output["macro_f1"], 3), 0.333)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["precision"], 0.0)
        self.assertEqual(
            round(output["per_class_confusion_matrix"]["b"]["f1"], 3), 0.667
        )

    def test_bad_evaluation_type(self):
        with self.assertRaises(ValueError):
            gold = []
            pred = []
            evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
            evaluator.run(evaluation_type="This is a bad evaluation type! :)")

    def test_multilabel_multiclass_evaluation_detailed_metrics(self):
        def find_label_func(*args):
            raise NotImplementedError("func not implemented")

        def find_label_data_func(*args):
            raise NotImplementedError("func not implemented")

        gold_annos = [["Tom Anniston", "Acme", "USA"]]
        pred_annos = [["Tom", "Acme", "USA since"]]
        evaluator = caikit.core.toolkit.QualityEvaluator(gold_annos, pred_annos)
        output = evaluator.run(
            caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS,
            find_label_func,
            find_label_data_func,
            detailed_metrics=True,
        )
        output = output["detailed_metrics"]
        self.assertEqual(round(output[0]["exact_match_precision"], 3), 0.333)
        self.assertEqual(round(output[0]["exact_match_recall"], 3), 0.333)
        self.assertEqual(round(output[0]["exact_match_f1"], 3), 0.333)
        self.assertEqual(output[0]["partial_match_precision"], 1)
        self.assertEqual(output[0]["partial_match_recall"], 1)
        self.assertEqual(output[0]["partial_match_f1"], 1)

    def test_find_partial_matches_basic_functionality(self):
        # When gold annotations are part of a prediction or vice versa, both annotations are counted as matches
        gold_annos = ["Tom Anniston", "Acme", "USA", "works"]
        pred_annos = ["Tom", "Acme", "USA since", "works at", "works at Acme"]

        expected_gold_matches = ["Tom Anniston", "Acme", "USA", "works"]
        expected_partial_pred_matches = [
            "Acme",
            "Tom",
            "USA since",
            "works at",
            "works at Acme",
        ]

        (
            actual_gold_matches,
            actual_pred_matches,
        ) = QualityEvaluator.find_partial_matches(gold_annos, pred_annos)
        self.assertEqual(
            sorted(actual_pred_matches), sorted(expected_partial_pred_matches)
        )
        self.assertEqual(sorted(actual_gold_matches), sorted(expected_gold_matches))

    def test_find_partial_matches_no_duplicates(self):
        # When a gold annotation is part of a prediction and prediction is also part of another gold annotation,
        # the same prediction should be counted only once, but both gold annotations are counted
        gold_annos = ["gian franco kasper", "ioc representative gian franco kasper"]
        pred_annos = ["representative gian franco kasper"]

        expected_gold_matches = [
            "gian franco kasper",
            "ioc representative gian franco kasper",
        ]
        expected_partial_pred_matches = ["representative gian franco kasper"]

        (
            actual_gold_matches,
            actual_pred_matches,
        ) = QualityEvaluator.find_partial_matches(gold_annos, pred_annos)
        self.assertEqual(
            sorted(actual_pred_matches), sorted(expected_partial_pred_matches)
        )
        self.assertEqual(sorted(actual_gold_matches), sorted(expected_gold_matches))

    def test_find_partial_matches_no_overlaps(self):
        # Overlaps are not matched, only subsets
        gold_annos = ["ioc representative gian franco kasper"]
        pred_annos = ["FIS president gian franco kasper"]

        expected_gold_matches = []
        expected_partial_pred_matches = []

        (
            actual_gold_matches,
            actual_pred_matches,
        ) = QualityEvaluator.find_partial_matches(gold_annos, pred_annos)
        self.assertEqual(
            sorted(actual_pred_matches), sorted(expected_partial_pred_matches)
        )
        self.assertEqual(sorted(actual_gold_matches), sorted(expected_gold_matches))

    def test_find_partial_matches_with_exact_match(self):
        # When an exact match is also part of another annotation, the other annotation should also be counted
        gold_annos = ["Acme", "Acme marketplace", "Bar"]
        pred_annos = ["Acme", "Foo", "Bar Giant"]

        expected_gold_matches = ["Acme", "Acme marketplace", "Bar"]
        expected_partial_pred_matches = ["Acme", "Bar Giant"]

        (
            actual_gold_matches,
            actual_pred_matches,
        ) = QualityEvaluator.find_partial_matches(gold_annos, pred_annos)
        self.assertEqual(
            sorted(actual_pred_matches), sorted(expected_partial_pred_matches)
        )
        self.assertEqual(sorted(actual_gold_matches), sorted(expected_gold_matches))

    def test_calc_f1_score(self):
        gold = ["a", "b", "c", "d"]
        pred = ["a", "b", "d", "e", "f"]
        precision, recall, f1_score = QualityEvaluator.calc_f1_score(gold, pred)
        expected_precision = 3 / 5
        expected_recall = 3 / 4

        self.assertEqual(precision, expected_precision)
        self.assertEqual(recall, expected_recall)
        self.assertEqual(round(f1_score, 3), 0.667)

    def test_calc_f1_score_error_partial(self):
        gold = ["a", "b", "c", "d"]
        pred = ["a", "b", "d", "e", "f"]

        def partial_match(gold, pred):
            return None

        with self.assertRaises(ValueError):
            precision, recall, f1_score = QualityEvaluator.calc_f1_score(
                gold, pred, partial_match
            )

    def test_calc_f1_score_partial(self):
        gold = ["a", "b", "c", "d"]
        pred = ["a", "b", "d", "e", "f"]

        def partial_match(gold, pred):
            # Returns all gold and predictions as matched
            return gold, pred

        precision, recall, f1_score = QualityEvaluator.calc_f1_score(
            gold, pred, partial_match
        )
        expected_precision = 5 / 5
        expected_recall = 4 / 4

        self.assertEqual(precision, expected_precision)
        self.assertEqual(recall, expected_recall)
        self.assertEqual(round(f1_score, 3), 1.0)

    def test_multilabel_multiclass_evaluation(self):
        gold = [["a", "b", "c"], ["a"]]
        pred = [["b", "c"], ["b"]]

        def find_label_func(ele):
            return ele

        def find_label_data_func(data, label):
            return [ele for ele in data if ele == label]

        evaluation_type = caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        test_labels = ["a", "b"]
        output = evaluator.run(
            evaluation_type, find_label_func, find_label_data_func, labels=test_labels
        )

        # checking if only the labels we passed in are in the result
        check = [
            True
            for label in output["per_class_confusion_matrix"]
            if label in test_labels
        ]
        if False in check:
            check = False
        else:
            check = True

        self.assertTrue(check)

        self.assertEqual(output["per_class_confusion_matrix"]["b"]["precision"], 0.5)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["recall"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["precision"], 0)
        self.assertEqual(round(output["macro_precision"], 3), 0.25)
        self.assertEqual(round(output["macro_recall"], 3), 0.5)
        self.assertEqual(round(output["macro_f1"], 3), 0.333)
        self.assertEqual(round(output["micro_precision"], 3), 0.5)
        self.assertEqual(round(output["micro_recall"], 3), 0.333)
        self.assertEqual(round(output["micro_f1"], 3), 0.4)
        self.assertEqual(output["micro_precision_partial_match"], 0)
        self.assertEqual(output["micro_recall_partial_match"], 0)
        self.assertEqual(output["micro_f1_partial_match"], 0)

    def test_multilabel_multiclass_evaluation_partial_match(self):
        gold = [["United States", "California", "Colorado"], ["Liberty"]]
        pred = [["States", "California"], ["Statue Liberty"]]

        def find_label_func(ele):
            return ele

        def find_label_data_func(data, label):
            return [ele for ele in data if ele == label]

        evaluation_type = caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        output = evaluator.run(
            evaluation_type,
            find_label_func,
            find_label_data_func,
            partial_match_metrics=True,
        )
        self.assertEqual(round(output["micro_precision_partial_match"], 3), 1)
        self.assertEqual(round(output["micro_recall_partial_match"], 3), 0.75)
        self.assertEqual(round(output["micro_f1_partial_match"], 3), 0.857)

    def test_singlelabel_multiclass_evaluation(self):
        gold = ["a", "b", "a", "b"]
        pred = ["b", "b", "b", "b"]
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        test_labels = ["b"]
        output = evaluator.run(
            evaluation_type=caikit.core.toolkit.EvalTypes.SINGLELABEL_MULTICLASS,
            labels=test_labels,
        )

        # checking if only the labels we passed in are in the result
        check = [
            True
            for label in output["per_class_confusion_matrix"]
            if label in test_labels
        ]
        if False in check:
            check = False
        else:
            check = True

        self.assertTrue(check)

        for label in test_labels:
            self.assertIn("true_positive", output["per_class_confusion_matrix"][label])
            self.assertIn("false_positive", output["per_class_confusion_matrix"][label])
            self.assertIn("false_negative", output["per_class_confusion_matrix"][label])
            self.assertIn("precision", output["per_class_confusion_matrix"][label])
            self.assertIn("recall", output["per_class_confusion_matrix"][label])
            self.assertIn("f1", output["per_class_confusion_matrix"][label])

        self.assertEqual(output["macro_precision"], 0.5)
        self.assertEqual(output["macro_recall"], 1.0)
        self.assertEqual(round(output["macro_f1"], 3), 0.667)
        self.assertEqual(output["micro_precision"], 0.5)
        self.assertEqual(output["micro_recall"], 1.0)
        self.assertEqual(round(output["micro_f1"], 3), 0.667)

    def test_multilabel_multiclass_evaluation_with_empty_labels(self):
        gold = [["a", "b"], ["b"], [], ["a"]]
        pred = [["a"], ["b"], ["a"], []]

        def find_label_func(ele):
            return ele

        def find_label_data_func(data, label):
            return [ele for ele in data if ele == label]

        evaluation_type = caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        output = evaluator.run(evaluation_type, find_label_func, find_label_data_func)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["precision"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["recall"], 0.5)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["precision"], 0.5)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["recall"], 0.5)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["true_positive"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["false_positive"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["a"]["false_negative"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["true_positive"], 1)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["false_positive"], 0)
        self.assertEqual(output["per_class_confusion_matrix"]["b"]["false_negative"], 1)
        self.assertEqual(round(output["macro_precision"], 3), 0.75)
        self.assertEqual(round(output["macro_recall"], 3), 0.5)
        self.assertEqual(round(output["macro_f1"], 3), 0.583)
        self.assertEqual(round(output["micro_precision"], 3), 0.667)
        self.assertEqual(round(output["micro_recall"], 3), 0.5)
        self.assertEqual(round(output["micro_f1"], 3), 0.571)
        self.assertEqual(round(output["micro_precision_partial_match"], 3), 0)
        self.assertEqual(round(output["micro_recall_partial_match"], 3), 0)
        self.assertEqual(round(output["micro_f1_partial_match"], 3), 0)

    def test_multilabel_multiclass_evaluation_with_no_find_label_func(self):
        gold = [["a", "b"], ["b"], [], ["a"]]
        pred = [["a"], ["b"], ["a"], []]

        def find_label_func(*_args, **_kwargs):
            raise NotImplementedError("Func not implemented")

        def find_label_data_func(*_args, **_kwargs):
            raise NotImplementedError("Func not implemented")

        evaluation_type = caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        output = evaluator.run(evaluation_type, find_label_func, find_label_data_func)
        # should calculate only micro metrics
        self.assertFalse(output["per_class_confusion_matrix"])
        self.assertEqual(output["macro_precision"], 0.0)
        self.assertEqual(output["macro_recall"], 0.0)
        self.assertEqual(output["macro_f1"], 0.0)
        self.assertEqual(round(output["micro_precision"], 3), 0.667)
        self.assertEqual(round(output["micro_recall"], 3), 0.5)
        self.assertEqual(round(output["micro_f1"], 3), 0.571)
        self.assertEqual(round(output["micro_precision_partial_match"], 3), 0)
        self.assertEqual(round(output["micro_recall_partial_match"], 3), 0)
        self.assertEqual(round(output["micro_f1_partial_match"], 3), 0)

    def test_multilabel_multiclass_evaluation_with_edge_case_zero_values(self):
        # true positives would be 0, resulting in 0 in denominator of F1 calculation
        gold = [["a", "b"], ["b"], [], ["a"]]
        pred = [[], ["a"], [], ["b"]]

        def find_label_func(*_args, **_kwargs):
            raise NotImplementedError("Func not implemented")

        def find_label_data_func(*_args, **_kwargs):
            raise NotImplementedError("Func not implemented")

        evaluation_type = caikit.core.toolkit.EvalTypes.MULTILABEL_MULTICLASS
        evaluator = caikit.core.toolkit.QualityEvaluator(gold, pred)
        output = evaluator.run(evaluation_type, find_label_func, find_label_data_func)
        # should calculate only micro metrics
        self.assertFalse(output["per_class_confusion_matrix"])
        self.assertEqual(output["macro_precision"], 0.0)
        self.assertEqual(output["macro_recall"], 0.0)
        self.assertEqual(output["macro_f1"], 0.0)
        self.assertEqual(round(output["micro_precision"], 3), 0.0)
        self.assertEqual(round(output["micro_recall"], 3), 0.0)
        self.assertEqual(round(output["micro_f1"], 3), 0.0)
        self.assertEqual(round(output["micro_precision_partial_match"], 3), 0)
        self.assertEqual(round(output["micro_recall_partial_match"], 3), 0)
        self.assertEqual(round(output["micro_f1_partial_match"], 3), 0)

    def test_calc_metrics_from_confusion_matrix(self):
        # gold = ['a', 'b', 'a', 'a']
        # pred = ['a', 'b', 'b' , 'b']

        confusion_matrix = {
            "a": F1Metrics(true_positive=1, false_positive=0, false_negative=2),
            "b": F1Metrics(true_positive=1, false_positive=2, false_negative=0),
        }

        metrics_summary = QualityEvaluator.calc_metrics_from_confusion_matrix(
            confusion_matrix
        )
        self.assertEqual(metrics_summary.per_class_confusion_matrix["a"].precision, 1)
        self.assertEqual(
            round(metrics_summary.per_class_confusion_matrix["a"].recall, 3), 0.333
        )
        self.assertEqual(
            round(metrics_summary.per_class_confusion_matrix["a"].f1, 3), 0.5
        )
        self.assertEqual(
            round(metrics_summary.per_class_confusion_matrix["b"].precision, 3), 0.333
        )
        self.assertEqual(round(metrics_summary.micro_metrics.precision, 3), 0.5)
        self.assertEqual(round(metrics_summary.micro_metrics.recall, 3), 0.5)
        self.assertEqual(round(metrics_summary.macro_metrics.precision, 3), 0.667)
        self.assertEqual(round(metrics_summary.macro_metrics.recall, 3), 0.667)
        self.assertEqual(round(metrics_summary.macro_metrics.f1, 3), 0.5)
        self.assertEqual(metrics_summary.micro_metrics.true_positive, 2)

    def test_convert_F1MetricsContainer_to_dict(self):
        confusion_matrix = {
            "a": F1Metrics(precision=0.5, recall=0.5, f1=0.5),
            "b": F1Metrics(precision=0.667, recall=0.667, f1=0.667),
        }

        micro_metrics = F1Metrics(
            precision=0.8,
            recall=0.5,
            f1=0.75,
            true_positive=5,
            false_negative=6,
            false_positive=7,
        )
        macro_metrics = F1Metrics(precision=0.5, recall=0.3, f1=0.4)
        metrics_summary = F1MetricsContainer(
            per_class_confusion_matrix=confusion_matrix,
            micro_metrics=micro_metrics,
            macro_metrics=macro_metrics,
        )
        metrics_out = QualityEvaluator.convert_F1MetricsContainer_to_dict(
            metrics_summary
        )
        self.assertEqual(
            metrics_out["per_class_confusion_matrix"]["a"]["precision"], 0.5
        )
        self.assertEqual(
            metrics_out["per_class_confusion_matrix"]["b"]["precision"], 0.667
        )
        self.assertEqual(round(metrics_out["micro_precision"], 3), 0.8)
        self.assertEqual(round(metrics_out["micro_f1"], 3), 0.75)
        self.assertEqual(round(metrics_out["macro_f1"], 3), 0.4)
        self.assertEqual(metrics_out["overall_tp"], 5)
        self.assertEqual(metrics_out["overall_fp"], 7)
        self.assertEqual(metrics_out["overall_fn"], 6)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for common enums functionality"""

# Standard
from enum import Enum
import json

# Third Party
from munch import Munch
import pytest
import yaml

# First Party
import py_to_proto

# Local
from caikit.core.data_model import enums

## Helpers #####################################################################


@pytest.fixture(autouse=True)
def reset_enums():
    """Helper to purge global attributes from the enums"""
    old_attrs = [attr for attr in dir(enums)]
    yield
    added_attrs = [attr for attr in dir(enums) if attr not in old_attrs]
    for added_attr in added_attrs:
        delattr(enums, added_attr)


## Tests #######################################################################


def test_import_enum_no_enum_class():
    """Make sure that importing an enum without an Enum class correctly creates
    one on the fly
    """

    class MyEnum(Enum):
        FOO = 1
        BAR = 2

    enum_name = MyEnum.__name__
    proto_enum = py_to_proto.descriptor_to_message_class(
        py_to_proto.dataclass_to_proto("foo", MyEnum)
    )

    assert not hasattr(enums, enum_name)
    enums.import_enum(proto_enum)
    assert hasattr(enums, enum_name)
    assert hasattr(enums, enum_name + "Rev")
    imported_enum = getattr(enums, enum_name)
    assert imported_enum.__name__ == enum_name
    assert issubclass(imported_enum, Enum)
    assert imported_enum.FOO.value == MyEnum.FOO.value
    assert imported_enum.BAR.value == MyEnum.BAR.value


def test_import_enum_with_enum_class():
    """Make sure that importing an enum with an Enum class does not create one
    on the fly
    """

    class MyEnum(Enum):
        FOO = 1
        BAR = 2

    enum_name = MyEnum.__name__
    proto_enum = py_to_proto.descriptor_to_message_class(
        py_to_proto.dataclass_to_proto("foo", MyEnum)
    )

    assert not hasattr(enums, enum_name)
    enums.import_enum(proto_enum, MyEnum)
    assert hasattr(enums, enum_name)
    assert hasattr(enums, enum_name + "Rev")
    imported_enum = getattr(enums, enum_name)
    assert imported_enum is MyEnum


def test_import_enum_extra_serialization():
    """Make sure that an imported enum gets correctly decorated with extra
    serialization methods
    """

    class MyEnum(Enum):
        FOO = 1
        BAR = 2

    enum_name = MyEnum.__name__
    proto_enum = py_to_proto.descriptor_to_message_class(
        py_to_proto.dataclass_to_proto("foo", MyEnum)
    )

    assert not hasattr(enums, enum_name)
    enums.import_enum(proto_enum, MyEnum)
    assert hasattr(enums, enum_name)
    assert hasattr(enums, enum_name + "Rev")

    # to_dict
    exp_dict = {"FOO": 1, "BAR": 2}
    assert MyEnum.to_dict() == exp_dict

    # to_munch
    munch_repr = MyEnum.to_munch()
    assert munch_repr == Munch(exp_dict)
    assert json.loads(munch_repr.toJSON()) == exp_dict
    assert yaml.safe_load(munch_repr.toYAML()) == exp_dict


def test_import_enum_rev():
    """Make sure that the reverse enum is created correctly"""

    class MyEnum(Enum):
        FOO = 1
        BAR = 2

    enum_name = MyEnum.__name__
    proto_enum = py_to_proto.descriptor_to_message_class(
        py_to_proto.dataclass_to_proto("foo", MyEnum)
    )
    assert not hasattr(enums, enum_name)
    enums.import_enum(proto_enum, MyEnum)
    assert hasattr(enums, enum_name)
    assert hasattr(enums, enum_name + "Rev")
    rev = getattr(enums, enum_name + "Rev")
    assert rev[1] == "FOO"
    assert rev[2] == "BAR"


def test_import_enum_invalid_proto_class():
    """Make sure that import_enum checks the proto class"""
    with pytest.raises(AttributeError):
        enums.import_enum("not a proto enum")


def test_import_enums_global():
    """Test that the import_enums utility pulls in all enums from the
    'current_globals' dict
    """

    class MyEnum(Enum):
        FOO = 1
        BAR = 2

    enum_name = MyEnum.__name__
    rev_name = enum_name + "Rev"
    proto_enum = py_to_proto.descriptor_to_message_class(
        py_to_proto.dataclass_to_proto("foo", MyEnum)
    )

    assert not hasattr(enums, enum_name)
    current_globals = {
        "protobufs": Munch(
            {
                "all_enum_names": [enum_name],
                enum_name: proto_enum,
            }
        )
    }
    enums.import_enums(current_globals)
    assert hasattr(enums, enum_name)
    assert hasattr(enums, rev_name)
    assert current_globals[enum_name] is getattr(enums, enum_name)
    assert current_globals[rev_name] is getattr(enums, rev_name)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Batcher transparently aggregates individual inference calls into unified
batches to call the run_batch implementation of the wrapped model.
"""

# Standard
from typing import Optional
import inspect
import queue
import threading

# First Party
import alog

# Local
from caikit.core import ModuleBase
from caikit.core.data_model.base import DataBase
from caikit.core.toolkit.wip_decorator import Action, WipCategory, work_in_progress

log = alog.use_channel("BATCHER")


# pylint: disable=too-many-instance-attributes
@work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
class Batcher:
    __doc__ = __doc__

    def __init__(
        self,
        model_name: str,
        model: ModuleBase,
        batch_size: int,
        batch_collect_delay_s: Optional[float] = None,
    ):
        """Configure the batcher to wrap the given model and run batches of at
        most the given size with the given delay to collect a batch.

        Args:
            model_name (str): The name of this model for error reporting
            model (ModuleBase): The model instance that will have inputs batched
            batch_size (int): The maximum size of a batch that will be sent to
                run_batch. Batches of smaller sizes will be sent if no
                additional requests are available.
            batch_collect_delay_s (Optional[float]): Number of seconds to wait
                for more requests to show up when filling the batch.
        """
        self._model_name = model_name
        self._model = model
        self._batch_size = batch_size
        self._batch_collect_delay_s = batch_collect_delay_s

        # The input queue and the output tasks dict are how the internal batch
        # thread manages receiving work and sharing results
        self._input_q = queue.SimpleQueue()
        self._finished_tasks = {}

        # The request number is used to generate an id for each request, but it
        # needs to be locked. We could use a uuid instead and not lock it, but
        # that would likely be slower (TODO: validate this assumption!)
        self._req_num = 0
        self._id_lock = threading.Lock()

        # These manage the lifecycle of the run thread. In order to avoid a busy
        # thread that never shuts down, the thread will self-terminate when
        # there is no more work ready
        self._ready_event = threading.Event()
        self._stop_event = threading.Event()
        self._batch_thread_start_lock = threading.Lock()
        self._batch_thread = None

        # Get a list of any/all default argument values for this model's run
        # function. This is needed to correctly fill in missing values when an
        # individual run() invocation is missing them, but others in the batch
        # have them.
        self._model_run_defaults = {
            param_name: param.default
            for param_name, param in inspect.signature(
                self._model.run
            ).parameters.items()
            if param.default
        }

    def __del__(self):
        """Shut down the internal thread"""
        self.stop()

    ## Public ##

    def run(self, **kwargs) -> DataBase:
        """This run function gives a facade to the underlying model's run
        function that is implemented by running batches of individual requests
        through the model's run_batch method.

        NOTE: Only kwargs accepted to simplify batching across inconsistent sets
            of kwargs (and only kwargs are used in the predict servicer)
        """
        # Make sure the batch thread is active
        self._ensure_batch_thread()

        # Get the ID for this request and make an event
        req_id = self._next_req_id()
        event = threading.Event()
        log.debug2("Starting request %d", req_id)

        # Put the args and kwargs onto the queue with the id and make sure that
        # the batch thread knows it's ready
        log.debug3("Queuing request %s", req_id)
        self._input_q.put((req_id, event, kwargs))
        self._ready_event.set()

        # Wait until notified
        log.debug3("Waiting for request %s", req_id)
        if not self._stop_event.is_set():
            event.wait()
        else:
            log.debug("Not waiting for event %s to complete", req_id)
            self._finished_tasks[req_id] = InterruptedError("Model stopped")
        log.debug3("Request %s finished", req_id)

        # Pull the results from _finished_tasks and handle errors
        res = self._finished_tasks.pop(req_id)
        if isinstance(res, Exception):
            raise RuntimeError(
                f"Exception caught for request {req_id} on model {self._model_name}"
            ) from res
        return res

    def stop(self):
        """Stop this batcher's run thread (cannot be undone)"""
        self._stop_event.set()
        with self._batch_thread_start_lock:
            if self._batch_thread and self._batch_thread.is_alive():
                # There is a race where the batch thread is waiting on the
                # ready_event to indicate that it has a first request so that it
                # doesn't preemptively shut down, so we need to unblock that
                # wait by setting the _ready_event.
                self._ready_event.set()
                if threading.current_thread() is not self._batch_thread:
                    self._batch_thread.join()
                log.debug2("Batch thread fully stopped")

                # Pull any remaining wait conditions off the input queue and
                # terminate them
                interrupt_err = InterruptedError("Model batcher stopped")
                while True:
                    try:
                        req_id, event, _ = self._input_q.get_nowait()
                        self._finished_tasks[req_id] = interrupt_err
                        event.set()
                        log.debug3("Stopped in-flight request [%s]", req_id)
                    except queue.Empty:
                        break
                log.debug2("Done stopping in-flight requests")

    ## Implementaiton Details ##

    def _ensure_batch_thread(self):
        """The run thread will stop itself if there's no work to do, so this
        function is called to ensure that it's up and running
        """
        # Once a model is stopped, it can't be restarted
        if self._stop_event.is_set():
            raise RuntimeError("Cannot restart a stopped model")

        if not self._batch_thread or not self._batch_thread.is_alive():
            with self._batch_thread_start_lock:
                # We only want to grab the lock if we're pretty sure we'll need
                # it, but there's a _chance_ that between the initial check and
                # getting the lock, someone else may have started the thread, so
                # we check again once we've got it.
                if not self._batch_thread or not self._batch_thread.is_alive():
                    log.debug3("Starting up batch thread")
                    self._batch_thread = threading.Thread(target=self._batch_thread_run)
                    self._ready_event.clear()
                    self._batch_thread.start()
                    log.debug3("Thread is started!")

    def _next_req_id(self):
        """Make a unique ID for this request"""
        with self._id_lock:
            self._req_num += 1
            return self._req_num

    def _batch_thread_run(self):
        """This function runs in an independent thread and manages pulling
        requests from the input queue, running the batch, and returning the
        completed results into _finished_tasks.
        """
        # Make sure the first input is available
        log.debug3("Batch thread started")
        self._ready_event.wait()
        log.debug3("Batch thread ready to work")

        # Start a fresh batch and keep the thread alive while there is anything
        # in the queue
        current_batch = []
        while True:
            # If the stop event is set, break out
            if self._stop_event.is_set():
                log.debug("Terminating batch thread for [%s]", self._model_name)
                err = InterruptedError("Model batcher stopped")
                for req_id, event, _ in current_batch:
                    log.debug2("Aborting batched request %s", req_id)
                    self._finished_tasks[req_id] = err
                    event.set()
                break

            # Look for more work and if found, add it to the batch
            no_more_work = False
            try:
                log.debug3(
                    "Getting next request. Waiting for %s", self._batch_collect_delay_s
                )
                next_req = self._input_q.get(
                    block=self._batch_collect_delay_s is not None,
                    timeout=self._batch_collect_delay_s,
                )
                log.debug2(
                    "Adding request %s to batch of size %d",
                    next_req[0],
                    len(current_batch),
                )
                current_batch.append(next_req)
            except queue.Empty:
                if not current_batch:
                    log.debug2(
                        "Shutting down active batch thread for %s", self._model_name
                    )
                    break
                no_more_work = True

            # If the batch is full, or the queue is empty, send the batch and
            # start a new one
            if len(current_batch) == self._batch_size or no_more_work:
                log.debug2("Running batch of size %d", len(current_batch))

                # Collect iterables for each kwarg, filling in gaps with their
                # defaults if possible
                batch_kwargs = {}
                invalid_reqs = {}
                for i, (_, _, req_kwargs) in enumerate(current_batch):
                    # Figure out kwarg names that are new to this batch and
                    # known and new kwargs that this req introduces to the batch
                    # pylint: disable=consider-iterating-dictionary
                    new_kwargs = [
                        kwarg_name
                        for kwarg_name in req_kwargs.keys()
                        if kwarg_name not in batch_kwargs
                    ]
                    missing_kwargs = [
                        kwarg_name
                        for kwarg_name in batch_kwargs.keys()
                        if kwarg_name not in req_kwargs
                    ]

                    # If there are new kwargs and the batch already has entries,
                    # we need to fill those entries in with the default value.
                    # This could cause us to realize that some of those previous
                    # requests were invalid because they were missing required
                    # kwargs!
                    if i > 0:
                        invalid_new_kwargs = [
                            kwarg
                            for kwarg in new_kwargs
                            if kwarg not in self._model_run_defaults
                        ]
                        if invalid_new_kwargs:
                            for j in range(i):
                                log.info(
                                    "<RUN71000210I>"
                                    "Rejecting invalid request to [%s] with "
                                    "missing required kwargs %s",
                                    self._model_name,
                                    invalid_new_kwargs,
                                )
                                invalid_reqs[j] = ValueError(
                                    f"Missing required arguments: {invalid_new_kwargs}"
                                )
                        for new_kwarg in new_kwargs:
                            batch_kwargs[new_kwarg] = [
                                self._model_run_defaults.get(new_kwarg)
                            ] * i

                    # If there are missing kwargs, fill them in with defaults.
                    # This could cause us to realize that this request is
                    # invalid!
                    missing_required_kwargs = [
                        kwarg
                        for kwarg in missing_kwargs
                        if kwarg not in self._model_run_defaults
                    ]
                    if missing_required_kwargs:
                        log.info(
                            "<RUN71000211I>",
                            "Rejecting invalid request to [%s] with missing required kwargs %s",
                            self._model_name,
                            missing_required_kwargs,
                        )
                        invalid_reqs[i] = ValueError(
                            f"Missing required arguments: {missing_required_kwargs}"
                        )
                    for kwarg in missing_kwargs:
                        # NOTE: Invalid requests will be removed later, but we
                        #   fill in the defaults for all requests to keep the
                        #   indices lined up.
                        dflt = self._model_run_defaults.get(kwarg)
                        log.debug3(
                            "Filling in default value for kwarg %s=%s", kwarg, dflt
                        )
                        req_kwargs[kwarg] = dflt

                    # Add to the batch kwargs, regardless of if it is valid so
                    # that the indices of the batch iterables can line up with
                    # the indices in the list of invalid requests
                    for kwarg_name, kwarg_val in req_kwargs.items():
                        batch_kwargs.setdefault(kwarg_name, []).append(kwarg_val)

                # Remove any entries from the batch that have been marked
                # invalid. Work backwards so that the earlier indices are still
                # valid once the later ones have been removed.
                for invalid_idx, invalid_err in sorted(
                    invalid_reqs.items(), reverse=True, key=lambda x: x[0]
                ):
                    # Slice the arg vals out of the kwarg lists
                    for kwarg_name, kwarg_vals in list(batch_kwargs.items()):
                        batch_kwargs[kwarg_name] = (
                            kwarg_vals[:invalid_idx] + kwarg_vals[invalid_idx + 1 :]
                        )

                    # Notify the waiting task and remove the entry from the
                    # batch so that it isn't processed later
                    req_id, event, _ = current_batch[invalid_idx]
                    current_batch = (
                        current_batch[:invalid_idx] + current_batch[invalid_idx + 1 :]
                    )
                    self._finished_tasks[req_id] = invalid_err
                    event.set()

                # Call the batch_run
                if current_batch:
                    try:
                        log.debug(
                            "Running with concrete batch size %d", len(current_batch)
                        )
                        log.debug4(batch_kwargs)
                        batch_res = self._model.run_batch(**batch_kwargs)
                        # pylint: disable=line-too-long
                        assert len(batch_res) == len(
                            current_batch
                        ), f"Got result of size [{len(batch_res)}] for batch of size [{len(current_batch)}]"
                        for i, (req_id, event, _) in enumerate(current_batch):
                            self._finished_tasks[req_id] = batch_res[i]
                            event.set()

                    # If an error occurred, push the single error as the result of
                    # all the requests in the batch
                    # pylint: disable=broad-exception-caught
                    except Exception as err:
                        log.error(
                            "Caught exception in batch thread for model %s: %s",
                            self._model_name,
                            err,
                        )
                        for req_id, event, _ in current_batch:
                            self._finished_tasks[req_id] = err
                            event.set()

                # Reset the next batch
                current_batch = []



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from unittest.mock import patch
import os

# Third Party
import pytest
import yaml

# First Party
import aconfig

# Local
from tests.conftest import temp_config
import caikit


# Let all of these tests call caikit.configure() without mucking the real config up
@pytest.fixture(autouse=True)
def patched_config():
    with temp_config({}):
        yield


CFG_1 = {"foo": 1, "combined": {"one": "bar"}}

CFG_2 = {"foo": 2, "combined": {"two": "baz"}}

CFG_3 = {"foo": 3, "combined": {"three": "buz"}}


def _dump_yml(cfg_dict: dict, path: str):
    with open(path, "w") as f:
        yaml.safe_dump(cfg_dict, f)
        f.flush()


def test_configure_raises_on_no_input():
    with pytest.raises(ValueError):
        caikit.configure()


def test_config_is_immutable():
    assert isinstance(caikit.get_config(), aconfig.ImmutableConfig)
    with pytest.raises(AttributeError):
        caikit.get_config().config_files = "foo"


def test_configure_reads_a_config_yml(tmp_path):
    path = os.path.join(os.path.join(tmp_path, "one.yml"))
    _dump_yml(CFG_1, path)
    caikit.configure(path)

    for k, v in CFG_1.items():
        assert caikit.get_config()[k] == v


def test_configure_reads_a_config_dict():
    caikit.configure(config_dict=CFG_1)

    for k, v in CFG_1.items():
        assert caikit.get_config()[k] == v


def test_configure_merges_over_existing_configs():
    caikit.configure(config_dict=CFG_1)

    caikit.configure(config_dict=CFG_2)

    cfg = caikit.get_config()
    # Not just the second one, we merged things in
    assert cfg != CFG_2
    # foo was overriden from two.yml
    assert cfg.foo == 2
    assert "one" in cfg.combined and "two" in cfg.combined


@patch.dict(os.environ, {"FOO": "42"})
def test_configure_picks_up_env_vars():
    caikit.configure(config_dict=CFG_1)

    # the FOO=42 env var is picked up
    assert caikit.get_config().foo == 42


def test_configure_adds_more_user_specified_files_from_env(tmp_path):
    path_2 = os.path.join(os.path.join(tmp_path, "two.yml"))
    _dump_yml(CFG_2, path_2)

    path_3 = os.path.join(os.path.join(tmp_path, "three.yml"))
    _dump_yml(CFG_3, path_3)

    with patch.dict(os.environ, {"CONFIG_FILES": f"{path_2},{path_3}"}):
        caikit.configure(config_dict=CFG_1)

    cfg = caikit.get_config()
    # all three configs applied, with CFG_3 applied last
    assert cfg.foo == 3
    assert "one" in cfg.combined and "two" in cfg.combined and "three" in cfg.combined


def test_configure_merges_lists():
    cfg1 = {"foo_list": [1, 2, 3]}
    cfg2 = {"foo_list": [4, 5, 6]}

    # Second list value is prepended to the first
    caikit.configure(config_dict=cfg1)
    caikit.configure(config_dict=cfg2)
    assert caikit.get_config().foo_list == [4, 5, 6, 1, 2, 3]

    # If values already existed in the list, they are popped then prepended
    cfg3 = {"foo_list": [1, 2]}
    caikit.configure(config_dict=cfg3)
    assert caikit.get_config().foo_list == [1, 2, 4, 5, 6, 1, 2, 3]


def test_merge_strategy():
    # If merge_strategy == 'override', then lists and dicts are not merged
    cfg1 = {"foo_list": [1, 2, 3], "foo_dict": {"bar": "baz"}}
    cfg2 = {
        "foo_list": [4, 5, 6],
        "foo_dict": {"baz": "bar"},
        "merge_strategy": "override",
    }

    caikit.configure(config_dict=cfg1)
    caikit.configure(config_dict=cfg2)
    assert caikit.get_config().foo_list == [4, 5, 6]
    assert caikit.get_config().foo_dict == {"baz": "bar"}


def test_merge_configs_aconfig():
    """Make sure merge_configs works as expected with aconfig.Config objects"""
    cfg1 = aconfig.Config({"foo": 1, "bar": {"baz": 2}, "bat": [1, 2, 3]})
    cfg2 = aconfig.Config({"foo": 11, "bar": {"buz": 22}, "bat": [5, 6, 7]})
    merged = caikit.config.config.merge_configs(cfg1, cfg2)
    assert merged == aconfig.Config(
        {
            "foo": 11,
            "bar": {"baz": 2, "buz": 22},
            "bat": [5, 6, 7, 1, 2, 3],
        }
    )


def test_merge_configs_dicts():
    """Make sure merge_configs works as expected with plain old dicts objects"""
    cfg1 = {"foo": 1, "bar": {"baz": 2}, "bat": [1, 2, 3]}
    cfg2 = {"foo": 11, "bar": {"buz": 22}, "bat": [5, 6, 7]}
    merged = caikit.config.config.merge_configs(cfg1, cfg2)
    assert merged == {
        "foo": 11,
        "bar": {"baz": 2, "buz": 22},
        "bat": [5, 6, 7, 1, 2, 3],
    }


def test_merge_configs_none_args():
    """Test that None args are handled correctly"""
    cfg1 = {"foo": 1, "bar": {"baz": 2}, "bat": [1, 2, 3]}
    cfg2 = {"foo": 11, "bar": {"buz": 22}, "bat": [5, 6, 7]}
    assert caikit.config.config.merge_configs(cfg1, None) == cfg1
    assert caikit.config.config.merge_configs(None, cfg2) == cfg2
    assert caikit.config.config.merge_configs(None, None) == {}



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module holds common utilities for managing timestamps in protobufs
"""
# Standard
import datetime

# Third Party
from google.protobuf import timestamp_pb2

# First Party
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("TIMESTAMPS")
error = error_handler.get(log)

TIMESTAMP_PROTOBUF_NAME = "google.protobuf.Timestamp"


def datetime_to_proto(datetime_: datetime.datetime) -> timestamp_pb2.Timestamp:
    """Builds a google.protobuf.timestamp_pb2.Timestamp out of the provided datetime.datetime"""
    error.type_check("<COR48155462E>", datetime.datetime, datetime_=datetime_)
    return timestamp_pb2.Timestamp(
        seconds=int(datetime_.timestamp()), nanos=int(datetime_.microsecond * 1e3)
    )


def proto_to_datetime(time_pb2: timestamp_pb2.Timestamp) -> datetime.datetime:
    """Builds a datetime.datetime out of the provided google.protobuf.timestamp_pb2.Timestamp"""
    try:
        error.type_check("<COR48166462E>", timestamp_pb2.Timestamp, time_pb2=time_pb2)
    except TypeError as err:
        # Compatibility for some python 3.8 / protobuf 3.x setups
        if "Timestamp" not in str(type(time_pb2)):
            raise err

    return datetime.datetime.fromtimestamp(time_pb2.seconds + (time_pb2.nanos / 1e9))



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard

# Standard
from contextlib import contextmanager
from os import path
from typing import Tuple
import asyncio
import subprocess
import tempfile
import time
import venv

# Third Party
import pytest

caikit_dir = path.abspath(path.join(path.dirname(__file__), "..", ".."))


async def waitForPort(port, timeout: int) -> bool:
    tmax = time.time() + timeout
    while time.time() < tmax:
        try:
            _, writer = await asyncio.wait_for(
                asyncio.open_connection("127.0.0.1", port), timeout=5
            )
            writer.close()
            await writer.wait_closed()
            return True
        except:
            await asyncio.sleep(1)
    return False


@contextmanager
def requirements(example: str) -> Tuple[str, str]:
    """Installs caikit and requirements.txt from relative path, returns python from venv
    and absolute path to the example.
    Raises subprocess.CalledProcessError if installation fails
    """
    with tempfile.TemporaryDirectory() as venv_dir:
        python = path.abspath(path.join(venv_dir, "bin", "python3"))
        pip = path.abspath(path.join(venv_dir, "bin", "pip3"))
        example_dir = path.abspath(path.join(caikit_dir, "examples", example))

        # Create a venv, install local version of caikit and local version of requirements
        venv.create(env_dir=venv_dir, system_site_packages=False, with_pip=True)
        try:
            subprocess.run([pip, "install", caikit_dir + "[all]"], check=True)
            subprocess.run(
                [
                    pip,
                    "install",
                    "-r",
                    path.join(example_dir, "requirements.txt"),
                ],
                check=True,
            )
            yield python, example_dir
        except subprocess.CalledProcessError as cpe:
            pytest.fail(
                "Could not install requirements for {}: {}".format(example, cpe)
            )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import os
import shutil
import tempfile

# Local
from caikit.core import toolkit

# Unit Test Infrastructure
from tests.base import TestCaseBase


class TestFeatureExtractor(TestCaseBase):
    def setUp(self):
        self.yaml_filename = os.path.join(
            self.fixtures_dir, "dummy_module", "config.yml"
        )
        self.json_filename = os.path.join(
            self.fixtures_dir, "dummy_module", "data.json"
        )
        self.pickle_filename = os.path.join(
            self.fixtures_dir, "dummy_module", "data.pkl"
        )
        self.csv_filename = os.path.join(self.fixtures_dir, "sample.csv")

    def test_load_yaml(self):
        yaml_dict = toolkit.load_yaml(self.yaml_filename)
        self.assertIsInstance(yaml_dict, dict)
        self.assertEqual(yaml_dict["train"]["batch_size"], 64)
        self.assertEqual(yaml_dict["utf_text"], "人工智能")

    def test_save_yaml(self):
        tempdir = tempfile.mkdtemp()
        tempyaml = os.path.join(tempdir, "temp.yml")

        try:
            toolkit.save_yaml(toolkit.load_yaml(self.yaml_filename), tempyaml)
            yaml_dict = toolkit.load_yaml(tempyaml)

        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

        self.assertIsInstance(yaml_dict, dict)
        self.assertEqual(yaml_dict["train"]["batch_size"], 64)
        self.assertEqual(yaml_dict["utf_text"], "人工智能")

    def test_load_json(self):
        json_dict = toolkit.load_json(self.json_filename)
        self.assertIsInstance(json_dict, dict)
        self.assertEqual(json_dict["tokens"][0]["begin"], 0)
        self.assertEqual(json_dict["utf_text"], "人工智能")

    def test_save_json(self):
        tempdir = tempfile.mkdtemp()
        tempjson = os.path.join(tempdir, "temp.json")

        try:
            toolkit.save_json(toolkit.load_json(self.json_filename), tempjson)
            json_dict = toolkit.load_json(tempjson)

        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

        self.assertEqual(json_dict["tokens"][0]["begin"], 0)
        self.assertEqual(json_dict["utf_text"], "人工智能")

    def test_load_csv(self):
        csv_list = toolkit.load_csv(self.csv_filename)
        self.assertEqual(csv_list[0], ["name", "age"])
        self.assertEqual(csv_list[1], ["Tom", "45"])
        self.assertEqual(csv_list[2], ["Jack", "24"])

    def test_save_csv(self):
        tempdir = tempfile.mkdtemp()
        tempcsv = os.path.join(tempdir, "temp.csv")
        try:
            toolkit.save_csv(toolkit.load_csv(self.csv_filename), tempcsv)
            csv_list = toolkit.load_csv(tempcsv)

        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

        self.assertEqual(csv_list[0], ["name", "age"])
        self.assertEqual(csv_list[1], ["Tom", "45"])
        self.assertEqual(csv_list[2], ["Jack", "24"])

    def test_load_dict_csv(self):
        csv_dict = toolkit.load_dict_csv(self.csv_filename)
        self.assertEqual(csv_dict[0]["name"], "Tom")
        self.assertEqual(csv_dict[0]["age"], "45")
        self.assertEqual(csv_dict[1]["name"], "Jack")
        self.assertEqual(csv_dict[1]["age"], "24")

    def test_save_dict_csv(self):
        tempdir = tempfile.mkdtemp()
        tempcsv = os.path.join(tempdir, "temp.csv")
        try:
            toolkit.save_dict_csv(
                toolkit.load_dict_csv(self.csv_filename), tempcsv, "w"
            )
            toolkit.save_dict_csv(
                toolkit.load_dict_csv(self.csv_filename), tempcsv, "a"
            )
            csv_dict = toolkit.load_dict_csv(tempcsv)

        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

        self.assertEqual(csv_dict[0]["name"], "Tom")
        self.assertEqual(csv_dict[0]["age"], "45")
        self.assertEqual(csv_dict[1]["name"], "Jack")
        self.assertEqual(csv_dict[1]["age"], "24")

    def test_load_pickle(self):
        data = toolkit.load_pickle(self.pickle_filename)
        self.assertEqual(data, [1, 2, 3])

    def test_save_pickle(self):
        tempdir = tempfile.mkdtemp()
        temppickle = os.path.join(tempdir, "temp.pkl")

        try:
            toolkit.save_pickle(toolkit.load_pickle(self.pickle_filename), temppickle)
            data = toolkit.load_pickle(temppickle)

        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

        self.assertEqual(data, [1, 2, 3])
